"""
Implementation of GIoU loss hybrid/combinations with other loss functions.
"""
from typing import Dict

import torch
import torchvision
from torch import nn
from torch.nn import functional as F


def bbox_xywh_to_xyxy(bbox: torch.Tensor, eps=1e-9) -> torch.Tensor:
    """
    Converts bbox format xywh to xyxy.

    Args:
        bbox: BBox in xywh format.
        eps: minimum width and height

    Returns:
        Bbox in xyxy format
    """
    bbox[..., 2] = torch.max(bbox[..., 0] + bbox[..., 2], bbox[..., 0] + eps)
    bbox[..., 3] = torch.max(bbox[..., 1] + bbox[..., 3], bbox[..., 1] + eps)
    return bbox


def bounded_giou(
    bboxes1: torch.Tensor,
    bboxes2: torch.Tensor,
    reduction: str = 'mean',
    is_xywh_format: bool = True,
    lower_bound: float = -1,
    upper_bound: float = 10,
    eps: float = 1e-9
) -> torch.Tensor:
    """
    Wrapper to torchvision GIoU loss. This loss can be unstable in case of weird bboxes dimensions (line).
    For that reasons lower and upper bound are introduced.
    Also, minimum weight and height are used (suppoerted by torchvision GIoU).

    Args:
        bboxes1: lhs bboxes
        bboxes2: rhs bboxes
        reduction: Loss batch reduction option (none, mean, sum)
        is_xywh_format: If bbox is in xywh format then it is converted to xyxy format
        lower_bound: Loss lower bound (clipped) - Makes loss more stable
        upper_bound: Loss upper bound (clipped) - Makes loss more stable
        eps: Minimum width/height - Makes loss more stable

    Returns:
        GIoU loss
    """
    if is_xywh_format:
        # transform to xyxy format
        bboxes1 = bbox_xywh_to_xyxy(bboxes1)
        bboxes2 = bbox_xywh_to_xyxy(bboxes2)

    giou = torchvision.ops.generalized_box_iou_loss(bboxes1, bboxes2, reduction='none', eps=eps)
    giou = torch.clamp(giou, min=lower_bound, max=upper_bound)
    if reduction == 'none':
        return giou
    if reduction == 'mean':
        return giou.mean()
    if reduction == 'sum':
        return giou.sum()

    raise AssertionError(f'Unknown reduction option "{reduction}"!')


class HybridL1GIoU(nn.Module):
    """
    Combines L1 loss with GIoU loss.
    """
    def __init__(
        self,
        w_l1: float = 5,
        w_giou: float = 2,
        reduction: str = 'mean',
        is_xywh_format: bool = True,
        giou_loss_lower_bound: float = -1,
        giou_loss_upper_bound: float = 10,
        eps: float = 1e-9,
        *args,
        **kwargs
    ):
        """
        Args:
            w_l1: L1 loss weight
            w_giou: GIoU loss weight
            reduction: Loss batch reduction option (none, mean, sum)
            is_xywh_format: If bbox is in xywh format then it is converted to xyxy format
            lower_bound: Loss lower bound (clipped) - Makes loss more stable
            upper_bound: Loss upper bound (clipped) - Makes loss more stable
            eps: Minimum width/height - Makes loss more stable
        """
        super().__init__(*args, **kwargs)
        self._w_l1 = w_l1
        self._w_giou = w_giou
        self._reduction = reduction
        self._is_xywh_format = is_xywh_format
        self._giou_loss_lower_bound = giou_loss_lower_bound
        self._giou_loss_upper_bound = giou_loss_upper_bound
        self._eps = eps

    def forward(self, bboxes1: torch.Tensor, bboxes2: torch.Tensor) -> Dict[str, torch.Tensor]:
        l1 = F.l1_loss(bboxes1, bboxes2, reduction=self._reduction)
        giou = bounded_giou(
            bboxes1=bboxes1,
            bboxes2=bboxes2,
            reduction=self._reduction,
            is_xywh_format=self._is_xywh_format,
            lower_bound=self._giou_loss_lower_bound,
            upper_bound=self._giou_loss_upper_bound,
            eps=self._eps
        )
        return {
            'loss': self._w_l1 * l1 + self._w_giou * giou,
            'l1': l1,
            'giou': giou
        }


class HybridGaussianNLLLossGIoU(nn.Module):
    """
    Combines GaussianNLLLoss loss with GIoU loss.
    """
    def __init__(
        self,
        w_nllloss: float = 4,
        w_giou: float = 1,
        reduction: str = 'mean',
        is_xywh_format: bool = True,
        giou_loss_lower_bound: float = -1,
        giou_loss_upper_bound: float = 10,
        eps: float = 1e-9,
        *args,
        **kwargs
    ):
        """
        Args:
            w_nllloss: GaussianNLLLoss loss weight
            w_giou: GIoU loss weight
            reduction: Loss batch reduction option (none, mean, sum)
            is_xywh_format: If bbox is in xywh format then it is converted to xyxy format
            lower_bound: Loss lower bound (clipped) - Makes loss more stable
            upper_bound: Loss upper bound (clipped) - Makes loss more stable
            eps: Minimum width/height - Makes loss more stable
        """
        super().__init__(*args, **kwargs)
        self._w_nllloss = w_nllloss
        self._w_giou = w_giou
        self._reduction = reduction
        self._is_xywh_format = is_xywh_format
        self._giou_loss_lower_bound = giou_loss_lower_bound
        self._giou_loss_upper_bound = giou_loss_upper_bound
        self._eps = eps

    def forward(self, mean: torch.Tensor, gt: torch.Tensor, var: torch.Tensor) -> Dict[str, torch.Tensor]:
        nllloss = F.gaussian_nll_loss(mean, gt, var, reduction=self._reduction)
        giou = bounded_giou(
            bboxes1=mean,
            bboxes2=gt,
            reduction=self._reduction,
            is_xywh_format=self._is_xywh_format,
            lower_bound=self._giou_loss_lower_bound,
            upper_bound=self._giou_loss_upper_bound,
            eps=self._eps
        )
        return {
            'loss': self._w_nllloss * nllloss + self._w_giou * giou,
            'nllloss': nllloss,
            'giou': giou
        }


def run_test() -> None:
    pred = torch.tensor([
        # [0.1, 0.1, 0.2, 0.2]
        [0.0000, 0.0625, 0.0000, 0.0000]
    ], dtype=torch.float32)
    gt = torch.tensor([
        # [0.1, 0.1, 0.1, 0.1]
        [0.0000, 0.0638, 0.0000, 0.0000]
    ], dtype=torch.float32)
    var = torch.tensor([
        [1.0, 1.0, 1.0, 1.0]
    ], dtype=torch.float32)

    loss = HybridL1GIoU()
    print(loss(pred, gt))
    loss = HybridGaussianNLLLossGIoU()
    print(loss(pred, gt, var))


if __name__ == '__main__':
    run_test()



"""
Dataset factory.
"""
from typing import Dict, Any, Optional, List

from motrack_motion.datasets.mot.core import MOTDataset
from motrack_motion.datasets.torch import TrajectoryDataset

DATASET_CATALOG = {
    'mot': MOTDataset
}


def dataset_factory(
    name: str,
    path: str,
    history_len: int,
    future_len: int,
    sequence_list: Optional[List[str]] = None,
    additional_params: Optional[Dict[str, Any]] = None
) -> TrajectoryDataset:
    """
    Creates dataset by given name.

    Args:
        name: Dataset name
        path: Dataset path
        history_len: Observed trajectory length
        future_len: Unobserved trajectory length
        sequence_list: Dataset split sequence list
        additional_params: Additional dataset parameters

    Returns:
        Initialized dataset
    """
    name = name.lower()

    additional_params = {} if additional_params is None else additional_params

    cls = DATASET_CATALOG[name]
    return cls(
        path=path,
        history_len=history_len,
        future_len=future_len,
        sequence_list=sequence_list,
        **additional_params
    )



"""
Torch dataset support.
Any Dataset that implements `TrajectoryDataset` interface can be used for training and evaluation.
"""
from abc import abstractmethod, ABC
from typing import Optional, Tuple, Dict, Any, List, Union

import numpy as np
import torch
from scipy.interpolate import CubicSpline
from torch.utils.data import Dataset

from motrack_motion.datasets import augmentations, transforms
from motrack_motion.datasets.common import BasicSceneInfo


def interpolate_by_fps(
    fps_multiplier: float,
    bboxes_obs: np.ndarray,
    bboxes_unobs: np.ndarray,
    ts_obs: np.ndarray,
    ts_unobs: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    if fps_multiplier == 1:
        return bboxes_obs, bboxes_unobs, ts_obs, ts_unobs

    ts = np.concatenate([ts_obs, ts_unobs]).reshape(-1)
    bboxes = np.concatenate([bboxes_obs, bboxes_unobs])

    # Base interval
    t_start, t_middle, t_end = [round(float(x)) for x in [ts[0], ts_obs[-1, 0], ts[-1]]]
    n_points = t_end - t_start + 1

    # Transformed interval
    if fps_multiplier > 1:
        assert isinstance(fps_multiplier, int) or fps_multiplier.is_integer()
        t_n_points = (round(fps_multiplier) - 1) * (n_points - 1) + n_points
        t_ts = np.linspace(t_start, t_end, num=t_n_points, endpoint=True, dtype=ts.dtype)
    else:
        step = round(1 / fps_multiplier)
        t_ts = np.arange(t_start, t_end, step, dtype=ts.dtype)

    t_bboxes = np.zeros(shape=(t_ts.shape[0], bboxes.shape[-1]), dtype=bboxes.dtype)
    for dim in range(bboxes.shape[-1]):
        cs = CubicSpline(ts, bboxes[:, dim])
        t_bboxes[:, dim] = cs(t_ts)

    # Create new inputs
    mask = (t_ts <= t_middle)
    new_ts_obs = t_ts[mask].reshape(-1, 1)
    new_ts_unobs = t_ts[~mask].reshape(-1, 1)
    new_bboxes_obs = t_bboxes[mask]
    new_bboxes_unobs = t_bboxes[~mask]

    return new_bboxes_obs, new_bboxes_unobs, new_ts_obs, new_ts_unobs


class TrajectoryDataset(ABC):
    """
    Defines interface for TrajectoryDataset.
    """
    def __init__(
        self,
        history_len: int,
        future_len: int,
        sequence_list: Optional[List[str]] = None
    ):
        self._history_len = history_len
        self._future_len = future_len
        self._sequence_list = sequence_list
    @abstractmethod
    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, Dict[str, Any]]:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @property
    @abstractmethod
    def scenes(self) -> List[str]:
        """
        Returns:
            List of scenes in dataset.
        """

    @abstractmethod
    def parse_object_id(self, object_id: str) -> Tuple[str, str]:
        """
        Parses and validates object id.

        Object id convention is `{scene_name}_{scene_object_id}` and is unique over all scenes.

        For MOT {scene_name} represents one video sequence.
        For SOT {scene_name} does not need to be unique for the sequence
        but `{scene_name}_{scene_object_id}` is always unique

        Args:
            object_id: Object id

        Returns:
            scene name, scene object id
        """

    @abstractmethod
    def get_object_category(self, object_id: str) -> str:
        """
        Gets category for object.

        Args:
            object_id: Object id

        Returns:
            Object category
        """

    @abstractmethod
    def get_scene_object_ids(self, scene_name: str) -> List[str]:
        """
        Gets object ids for given scene name

        Args:
            scene_name: Scene name

        Returns:
            Scene objects
        """

    def get_scene_number_of_object_ids(self, scene_name: str) -> int:
        """
        Gets number of unique objects in the scene.

        Args:
            scene_name: Scene name

        Returns:
            Number of objects in the scene
        """
        return len(self.get_scene_object_ids(scene_name))

    @abstractmethod
    def get_object_data_length(self, object_id: str) -> int:
        """
        Gets total number of data points for given `object_id` for .

        Args:
            object_id: Object id

        Returns:
            Number of data points
        """

    @abstractmethod
    def get_object_data_label(self, object_id: str, index: int, relative_bbox_coords: bool = True) -> dict:
        """
        Get object data point index.

        Args:
            object_id: Object id
            index: Index
            relative_bbox_coords: Scale bbox coords to [0, 1]

        Returns:
            Data point.
        """

    @abstractmethod
    def get_object_data_label_by_frame_index(
        self,
        object_id: str,
        frame_index: int,
        relative_bbox_coords: bool = True
    ) -> Optional[dict]:
        """
        Like `get_object_data_label` but data is relative to given frame_index.
        If object does not exist in given frame index then None is returned.

        Args:
            object_id: Object id
            frame_index: Frame Index
            relative_bbox_coords: Scale bbox coords to [0, 1]

        Returns:
            Data point.
        """

    @abstractmethod
    def get_scene_info(self, scene_name: str) -> BasicSceneInfo:
        """
        Get scene metadata by name.

        Args:
            scene_name: Scene name

        Returns:
            Scene metadata
        """

    @abstractmethod
    def get_scene_image_path(self, scene_name: str, frame_index: int) -> str:
        """
        Get image (frame) path for given scene and frame id.

        Args:
            scene_name: scene name
            frame_index: frame index

        Returns:
            Frame path
        """


class TorchTrajectoryDataset(Dataset):
    """
    PyTorch wrapper for Trajectory dataset
    """
    def __init__(
        self,
        dataset: TrajectoryDataset,
        transform: Optional[transforms.InvertibleTransform] = None,
        augmentation_before_transform: Optional[augmentations.TrajectoryAugmentation] = None,
        augmentation_after_transform: Optional[augmentations.TrajectoryAugmentation] = None,
        fps_multiplier: float = 1,
    ) -> None:
        """
        Args:
            dataset: Dataset that implements `__len__` and `__getitem`
            transform: Item postprocess
        """
        super().__init__()
        self._dataset = dataset
        self._fps_multiplier = fps_multiplier

        self._transform = transform
        if self._transform is None:
            self._transform = transforms.IdentityTransform()

        self._augmentation_before_transform = augmentation_before_transform
        if self._augmentation_before_transform is None:
            self._augmentation_before_transform = augmentations.IdentityAugmentation()

        self._augmentation_after_transform = augmentation_after_transform
        if self._augmentation_after_transform is None:
            self._augmentation_after_transform = augmentations.IdentityAugmentation()

    @property
    def dataset(self) -> TrajectoryDataset:
        """
        Returns: MOT core dataset class object.
        """
        return self._dataset

    def __len__(self) -> int:
        return len(self._dataset)

    def __getitem__(self, index: int) -> Dict[str, Union[dict, torch.Tensor]]:
        bboxes_obs, bboxes_unobs, ts_obs, ts_unobs, metadata = self._dataset[index]
        bboxes_obs, bboxes_unobs, ts_obs, ts_unobs = interpolate_by_fps(self._fps_multiplier, bboxes_obs, bboxes_unobs, ts_obs, ts_unobs)

        orig_bboxes_obs = torch.from_numpy(bboxes_obs)
        orig_bboxes_unobs = torch.from_numpy(bboxes_unobs)
        ts_obs = torch.from_numpy(ts_obs)
        ts_unobs = torch.from_numpy(ts_unobs)
        if 'images' in metadata:
            metadata['images'] = torch.from_numpy(metadata['images'])
        if 'flow' in metadata:
            metadata['flow'] = torch.from_numpy(metadata['flow'])

        # Trajectory transformations
        bboxes_obs, aug_bboxes_unobs, ts_obs, ts_unobs = \
            self._augmentation_before_transform(orig_bboxes_obs, orig_bboxes_unobs, ts_obs, ts_unobs)

        t_bboxes_obs, t_aug_bboxes_unobs, t_ts_obs, t_ts_unobs, _, _ = \
            self._transform([bboxes_obs, aug_bboxes_unobs, ts_obs, ts_unobs, metadata, None], shallow=False)

        _, t_bboxes_unobs, _, _, _, _ = \
            self._transform([bboxes_obs, orig_bboxes_unobs, ts_obs, ts_unobs, metadata, None], shallow=False)

        t_bboxes_obs, t_aug_bboxes_unobs, t_ts_obs, t_ts_unobs = \
            self._augmentation_after_transform(t_bboxes_obs, t_aug_bboxes_unobs, t_ts_obs, t_ts_unobs)

        return {
            't_bboxes_obs': t_bboxes_obs,
            't_aug_bboxes_unobs': t_aug_bboxes_unobs,
            't_ts_obs': t_ts_obs,
            't_ts_unobs': t_ts_unobs,
            'orig_bboxes_obs': orig_bboxes_obs,
            'orig_bboxes_unobs': orig_bboxes_unobs,
            't_bboxes_unobs': t_bboxes_unobs,
            'metadata': metadata
        }


def run_test() -> None:
    ts_obs = np.array([0, 1, 2, 3], dtype=np.float32).reshape(-1, 1)
    ts_unobs = np.array([4, 5], dtype=np.float32).reshape(-1, 1)
    bboxes_obs = np.array([[0, 0, 1, 1], [0.1, 0.1, 1.1, 1.1], [0.2, 0.2, 1.2, 1.2], [0.3, 0.3, 1.3, 1.3]], dtype=np.float32)
    bboxes_unobs = np.array([[0.4, 0.4, 1.4, 1.4], [0.5, 0.5, 1.5, 1.5]], dtype=np.float32)

    new_bboxes_obs, new_bboxes_unobs, new_ts_obs, new_ts_unobs = \
        interpolate_by_fps(1, bboxes_obs, bboxes_unobs, ts_obs, ts_unobs)

    print(f'Multiplier=1 (shape): {new_ts_obs.shape=}, {new_ts_unobs.shape=}, '
          f'{new_bboxes_obs.shape=}, {new_bboxes_unobs.shape=}')
    print(f'Multiplier=1: {new_ts_obs=}, {new_ts_unobs=}, {new_bboxes_obs=}, {new_bboxes_unobs=}')

    new_bboxes_obs, new_bboxes_unobs, new_ts_obs, new_ts_unobs = \
        interpolate_by_fps(2, bboxes_obs, bboxes_unobs, ts_obs, ts_unobs)

    print(f'Multiplier=2 (shape): {new_ts_obs.shape=}, {new_ts_unobs.shape=}, '
          f'{new_bboxes_obs.shape=}, {new_bboxes_unobs.shape=}')
    print(f'Multiplier=2: {new_ts_obs=}, {new_ts_unobs=}, {new_bboxes_obs=}, {new_bboxes_unobs=}')

    new_bboxes_obs, new_bboxes_unobs, new_ts_obs, new_ts_unobs = \
        interpolate_by_fps(0.5, bboxes_obs, bboxes_unobs, ts_obs, ts_unobs)

    print(f'Multiplier=0.5 (shape): {new_ts_obs.shape=}, {new_ts_unobs.shape=}, '
          f'{new_bboxes_obs.shape=}, {new_bboxes_unobs.shape=}')
    print(f'Multiplier=0.5: {new_ts_obs=}, {new_ts_unobs=}, {new_bboxes_obs=}, {new_bboxes_unobs=}')

if __name__ == '__main__':
    run_test()



"""
Implementations of data transformations.
"""
from abc import ABC, abstractmethod
from typing import Collection, Union, Optional

import torch

TensorCollection = Union[torch.Tensor, Collection[torch.Tensor], dict, None]


class Transform(ABC):
    """
    Maps data with implemented transformation.
    """
    def __init__(self, name: str):
        """
        Args:
            name: Transformation name.
        """
        self._name = name

    @property
    def name(self) -> str:
        """
        Returns:
            Transformation name
        """
        return self._name

    @abstractmethod
    def apply(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        """
        Perform transformation on given raw data.

        Args:
            data: Raw data
            shallow: Take shallow copy of data (may cause side effects but faster in general)

        Returns:
            Transformed data
        """
        pass

    def __call__(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        return self.apply(data, shallow=shallow)


class InvertibleTransform(Transform, ABC):
    """
    Transform that also implements `inverse` method.
    """
    def __init__(self, name: str):
        super().__init__(name=name)

    @abstractmethod
    def inverse(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        """
        Performs inverse transformation on given transformed data.

        Args:
            data: Transformed data
            shallow: Take shallow copy of data (may cause side effects)

        Returns:
            "Untransformed" data
        """
        pass


class InvertibleTransformWithVariance(InvertibleTransform):
    """
    Extended InvertibleTransform with inverse for std/var
    """
    def __init__(self, name: str):
        super().__init__(name=name)

    @abstractmethod
    def inverse_std(
        self,
        t_std: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        """
        Performs "inverse" transformation on std given the transformed data.

        Args:
            t_std: Transformed std
            additional_data: Data required to perform inverse std operation
            shallow: Take shallow copy of data (may cause side effects)

        Returns:
            "Untransformed" std
        """
        pass

    @abstractmethod
    def inverse_var(
        self,
        t_var: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        """
        Performs "inverse" transformation on variance given the transformed data.

        Args:
            t_var: Transformed variance
            additional_data: Data required to perform inverse std operation
            shallow: Take shallow copy of data (may cause side effects)

        Returns:
            "Untransformed" variance
        """
        pass


class IdentityTransform(InvertibleTransformWithVariance):
    """
    Transformation neutral operator.
    """
    def __init__(self):
        super().__init__(name='identity')

    def apply(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        return data

    def inverse(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        return data

    def inverse_std(self, t_std: torch.Tensor, additional_data: Optional[TensorCollection] = None,
                    shallow: bool = True) -> TensorCollection:
        return t_std

    def inverse_var(self, t_var: torch.Tensor, additional_data: Optional[TensorCollection] = None,
                    shallow: bool = True) -> TensorCollection:
        return t_var



"""
Implementation of training augmentations
"""
from motrack_motion.datasets.augmentations.trajectory import (
    TrajectoryAugmentation,
    IdentityAugmentation,
    GaussianNoiseAugmentation,
    CompositionAugmentation,
    create_identity_augmentation_config
)



"""
Model factory method
"""
import enum
from typing import Union, Optional

from motrack_motion.datasets.transforms import InvertibleTransform, InvertibleTransformWithVariance
from motrack_motion.models.train.base import LightningTrainConfig, LightningModuleBase
from motrack_motion.models.train.end_to_end import EndToEndFilterTrainingModule
from motrack_motion.models.architectures.factory import ModelType, model_factory


class TrainingModuleType(enum.Enum):
    """
    Enumerated implemented training modules
    """
    END_TO_END = 'end_to_end'


    @classmethod
    def from_str(cls, value: str) -> 'TrainingModuleType':
        for v in cls:
            if v.value.lower() == value.lower():
                return v

        raise ValueError(f'Can\'t create ModelType from "{value}". Possible values: {list(cls)}')


def load_or_create_training_module(
    train_module_type: Union[TrainingModuleType, str],
    train_module_params: dict,
    model_type: Union[ModelType, str],
    model_params: dict,
    checkpoint_path: Optional[str] = None,
    n_train_steps: Optional[int] = None,
    train_params: Optional[dict] = None,
    transform_func: Optional[Union[InvertibleTransform, InvertibleTransformWithVariance]] = None
) -> LightningModuleBase:
    """
    Loads trained (if given checkpoint path) or creates new model given name and parameters.
    If model is trainable (check ModelType) then it can use train config. Otherwise, it can be loaded from checkpoint.
    Parameters combinations:
    - checkpoint_path is None and train_params is None - not allowed
    - checkpoint_path is None and train_params is not None - model is used for training from scratch
    - checkpoint_path is not None and train_params is None - model is used for inference
    - checkpoint_path is None and train_params is None - model is used for training from checkpoint (continue training)
    Args:
        train_module_type: Module type
        train_module_params: Module parameters
        model_type: Model type
        model_params: Model parameters
        checkpoint_path: Load pretrained model
        n_train_steps: Number of train steps
        train_params: Parameters for model training
        transform_func: Transform function (applied before loss)
    Returns:
        Model
    """
    if isinstance(train_module_type, str):
        train_module_type = TrainingModuleType.from_str(train_module_type)

    catalog = {
        TrainingModuleType.END_TO_END: EndToEndFilterTrainingModule
    }

    # Load model
    model = model_factory(model_type, model_params)

    if n_train_steps is not None:
        # Override train steps
        train_params['n_train_steps'] = n_train_steps

    # Create training config
    train_config = LightningTrainConfig(**train_params) if train_params is not None else None
    if checkpoint_path is None and train_config is None:
        raise ValueError('Train config and checkpoint path can\'t be both None!')

    # Create training module
    module_cls = catalog[train_module_type]
    if checkpoint_path is not None:
        return module_cls.load_from_checkpoint(
            **train_module_params,
            model=model,
            checkpoint_path=checkpoint_path,
            train_config=train_config,
            transform_func=transform_func
        )

    # noinspection PyTypeChecker
    return module_cls(
        **train_module_params,
        model=model,
        train_config=train_config,
        transform_func=transform_func
    )



"""
Pipeline (tools scripts) functions.
"""
import logging
import os
import time
from datetime import datetime
from pathlib import Path
from typing import Callable

from omegaconf import DictConfig, OmegaConf

from motrack_motion.common import project, formats, conventions
from motrack_motion.utils import rich

logger = logging.getLogger('PipelineUtils')


def task(task_name: str) -> Callable:
    """
    Optional decorator that wraps the task function in extra utilities.

    Args:
        task_name: Task name

    Returns:
        Task wrapper decorator
    """
    def task_wrapper(task_func: Callable) -> Callable:
        """
        Args:
            task_func: Function to wrap

        Returns:
            Wrapped function
        """

        def wrap(cfg: DictConfig):
            # Extracting `master_path` from optional `cfg.path.master`.
            master_path = None
            paths = cfg.get('path')
            if paths is not None:
                master_path = paths.get('master')
            master_path = project.OUTPUTS_PATH if master_path is None else master_path
            Path(master_path).mkdir(parents=True, exist_ok=True)

            # Print config
            rich.print_config_tree(cfg, resolve=True, save_to_file=True)

            # Store config history
            store_run_history_config(master_path, cfg, task_name=task_name)

            # execute the task
            start_time = time.time()
            # Parse config
            cfg = OmegaConf.to_object(cfg)

            # Run
            task_func(cfg=cfg)
            logger.info(f"'{task_func.__name__}' execution time: {time.time() - start_time} (s)")

        return wrap

    return task_wrapper


def store_run_history_config(output_dir: str, cfg: DictConfig, task_name: str) -> None:
    """
    Stores run config of the task run.

    Args:
        output_dir: Task output path
        cfg: Task config
        task_name: Task name
    """
    experiment_path = conventions.get_experiment_path(output_dir, cfg.dataset.name, cfg.experiment)
    config_dirpath = conventions.get_run_history_path(experiment_path)
    dt = datetime.now().strftime(formats.DATETIME_FORMAT)
    config_path = os.path.join(config_dirpath, f'{dt}_{task_name}.yaml')
    Path(config_dirpath).mkdir(parents=True, exist_ok=True)
    with open(config_path, 'w', encoding='utf-8') as f:
        f.write(OmegaConf.to_yaml(cfg))



"""
PyTorch's extension with simple functions
"""
from torch.optim import Optimizer


def get_optim_lr(optimizer: Optimizer) -> float:
    """
    Gets current optimizer learning rate

    Args:
        optimizer: Torch optimizer

    Returns:
        Learning rate
    """
    for param_group in optimizer.param_groups:
        return param_group['lr']



"""
Dataset utils
"""
from collections import defaultdict
from typing import List, Optional, Dict, Union

import numpy as np
import torch
from torch.utils.data import default_collate

from motrack_motion.datasets.augmentations.trajectory import TrajectoryAugmentation, IdentityAugmentation


class TrajectoryDataloaderCollateFunctional:
    """
    Trajectory Dataloader collate func wrapper that supports configurable augmentations.
    """
    def __init__(self, augmentation: Optional[TrajectoryAugmentation] = None):
        """
        Creates `trajectory_dataloader_collate_func` wrapper by adding the augmentations after collate function

        Args:
            augmentation: Augmentation applied after collate function is applied (optional)

        Returns:
            `trajectory_dataloader_collate_func` with (optional) augmentations.
        """
        self._augmentation = augmentation
        if self._augmentation is None:
            self._augmentation = IdentityAugmentation()

    def __call__(self, items: List[Dict[str, Union[dict, torch.Tensor]]]) \
            -> Dict[str, Union[dict, torch.Tensor]]:
        """
        Trajectory collate func: Standard way to batch sequences of dimension (T, *shape)
        where T is time dimension and shape is feature dimension is to create batch
        of size (B, T, *shape) but for Time series it makes more sense to do it as (T, B, *shape)
        which requires custom collate_func

        Args:
            items: Items gathered from WeatherDataset

        Returns:
            collated tensors
        """
        unstack_items: Dict[str, List[Union[dict, torch.Tensor]]] = defaultdict(list)
        for item in items:
            for k, v in item.items():
                unstack_items[k].append(v)
        collated_batch = {}
        for k, v in unstack_items.items():
            if k != 'metadata':
                collated_batch[k] = torch.stack(v, dim=1)
            else:
                collated_batch[k] = {}
                for m_k in v[0].keys():
                    if m_k in ['flow', 'images']:
                        collated_batch[k][m_k] = torch.stack([item[m_k] for item in v], dim=1)
                    else:
                        collated_batch[k][m_k] = default_collate([item[m_k] for item in v])

        x_obs, x_aug_unobs, t_obs, t_unobs = \
            [collated_batch[k] for k in ['t_bboxes_obs', 't_aug_bboxes_unobs', 't_ts_obs', 't_ts_unobs']]

        # Apply augmentations at batch level (optional)
        x_obs, x_aug_unobs, t_obs, t_unobs = self._augmentation(x_obs, x_aug_unobs, t_obs, t_unobs)
        collated_batch['t_bboxes_obs'] = x_obs
        collated_batch['t_aug_bboxes_unobs'] = x_aug_unobs
        collated_batch['t_ts_obs'] = t_obs
        collated_batch['t_ts_unobs'] = t_unobs

        return collated_batch


def split_trajectory_observed_unobserved(frame_ids: List[int], bboxes: np.ndarray, history_len: int):
    """
    Splits trajectory time points and bboxes int observed (input) trajectory
    and unobserved (ground truth) trajectory.

    Args:
        frame_ids: Full trajectory frame ids
        bboxes: Full trajectory bboxes
        history_len: Observed trajectory length

    Returns:
        - Observed trajectory bboxes
        - Unobserved trajectory bboxes
        - Observed trajectory time points
        - Unobserved trajectory time points
    """

    # Time points
    frame_ts = np.array(frame_ids, dtype=np.float32)
    frame_ts = frame_ts - frame_ts[0] + 1  # Transforming to relative time values
    frame_ts = np.expand_dims(frame_ts, -1)

    # Observed - Unobserved
    bboxes_obs = bboxes[:history_len]
    bboxes_unobs = bboxes[history_len:]
    frame_ts_obs = frame_ts[:history_len]
    frame_ts_unobs = frame_ts[history_len:]

    return bboxes_obs, bboxes_unobs, frame_ts_obs, frame_ts_unobs



"""
End-to-End filter interface.
"""
from abc import ABC, abstractmethod
from typing import Union, Tuple, List

import torch
from torch import nn


class FilterModule(nn.Module, ABC):
    """
    Interface for the filter module.
    """
    def __init__(self, bounded_variance: bool = False, bounded_value: float = 0.01, *args, **kwargs):
        """
        Args:
            bounded_variance: Use soft-plus instead exp for variance post-process and bound it with `bounded_value` as lower value
            bounded_value: Bounded variance value
        """
        super().__init__(*args, **kwargs)
        self._bounded_variance = bounded_variance
        self._bounded_value = bounded_value

    @abstractmethod
    def initialize(self, batch_size: int, device: Union[str, torch.device]) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Creates initial filter state before any value is observed.

        Args:
            batch_size: Batch size
            device: Device to place initial state

        Returns:
            Initial state (hidden, time)
        """

    @abstractmethod
    def estimate_prior(self, z0: torch.Tensor, t0: torch.Tensor, t1: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Estimates the next prior state at time t1 given the posterior (z0, t0)
        Args:
            z0: Previous posterior latent state
            t0: Previous posterior time
            t1: Next prior time

        Returns:
            - Estimated prior mean
            - Estimated prior variance
            - Prior latent state
        """

    @abstractmethod
    def estimate_posterior(self, z1_prior: torch.Tensor, z1_evidence: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Estimates the next posterior state given the current prior state (z1_prior)
        and the observed latent evidence (z1_evidence)

        Args:
            z1_prior: Latent prior state
            z1_evidence: Latent evidence

        Returns:
            - Estimated posterior mean
            - Estimated posterior variance
            - Posterior latent state
        """

    @abstractmethod
    def encode_obs(self, x: torch.Tensor) -> torch.Tensor:
        """
        Encodes observation into latent state.

        Args:
            x: Raw observation

        Returns:
            Latent observation
        """

    @abstractmethod
    def encode_unobs(self, x: torch.Tensor) -> torch.Tensor:
        """
        Encodes "unobserved" observation into latent state.

        Args:
            x: Raw "unobserved" observation

        Returns:
            Latent "unobserved" observation
        """

    def next(self, z0: torch.Tensor, z1_evidence: torch.Tensor, t0: torch.Tensor, t1: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Combines `estimate_prior` and `estimate_posterior` into a single function. Convention for training.

        Args:
            z0: Previous posterior latent state
            z1_evidence: Latent evidence
            t0: Previous posterior time
            t1: Next prior time

        Returns:
            - Estimated prior mean
            - Estimated prior variance
            - Estimated posterior mean
            - Estimated posterior variance
            - Posterior latent state
        """
        x1_prior_mean, x1_prior_log_var, z1_prior = self.estimate_prior(z0, t0, t1)
        x1_posterior_mean, x1_posterior_log_var, z1_posterior = self.estimate_posterior(z1_prior, z1_evidence)
        return x1_prior_mean, x1_prior_log_var, x1_posterior_mean, x1_posterior_log_var, z1_posterior

    def encode_obs_trajectory(self, x_obs: torch.Tensor, t_obs: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Encoded full observed trajectory into latent state.

        Args:
            x_obs: Trajectory features
            t_obs: Trajectory time points

        Returns:
            - Trajectory latent summarization
            - Last time point
        """
        n_warmup_steps, batch_size, _ = t_obs.shape

        z0, t0 = self.initialize(batch_size, t_obs.device)
        z1_obs = self.encode_obs(x_obs)
        for i in range(n_warmup_steps):
            t1 = t_obs[i]
            z1 = z1_obs[i]
            _, _, _, _, z0 = self.next(z0, z1, t0, t1)
            t0 = t1

        return z0, t0

    def forward(self, x_obs: torch.Tensor, t_obs: torch.Tensor, x_unobs: torch.Tensor, t_unobs: torch.Tensor, mask: Union[List[bool], bool] = True) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Performs full forward step:
        - Encodes full observed (warm-up) trajectory into latent state
        - Filters full trajectory (including observed and unobserved)

        Args:
            x_obs: Observed trajectory (warm-up)
            t_obs: Observed trajectory time-points
            x_unobs: Unobserved trajectory (to be filtered)
            t_unobs: Unobserved trajectory time points
            mask: Simulates missing "unobserved" tr

        Returns:

        """
        if isinstance(mask, bool):
            mask = [mask for _ in range(t_unobs.shape[0])]
        assert len(mask) == t_unobs.shape[0], f'Invalid mask length {len(mask)}. Expected: {t_unobs.shape[0]}.'

        z0, t0 = self.encode_obs_trajectory(x_obs, t_obs)

        n_estimation_steps, _, _ = t_unobs.shape
        priors_mean, priors_log_var, posteriors_mean, posteriors_log_var = [], [], [], []
        z1_unobs = self.encode_unobs(x_unobs)
        i_lag = 0  # measurement index lags in case of missing points
        for i in range(n_estimation_steps):
            t1 = t_unobs[i]
            if mask[i]:
                # Data is present - Estimate the prior and the posterior
                z1 = z1_unobs[i - i_lag]
                x1_prior_mean, x1_prior_log_var, x1_posterior_mean, x1_posterior_log_var, z0 = self.next(z0, z1, t0, t1)
                posteriors_mean.append(x1_posterior_mean)
                posteriors_log_var.append(x1_posterior_log_var)
            else:
                # Data is not present - Just estimate the prior
                x1_prior_mean, x1_prior_log_var, z0 = self.estimate_prior(z0, t0, t1)
                i_lag += 1

            t0 = t1

            priors_mean.append(x1_prior_mean)
            priors_log_var.append(x1_prior_log_var)

        priors_mean, priors_log_var, posteriors_mean, posteriors_log_var = \
            [(torch.stack(v) if len(v) > 0 else torch.empty(0, dtype=torch.float32))
             for v in [priors_mean, priors_log_var, posteriors_mean, posteriors_log_var]]

        assert z0 is not None
        return priors_mean, priors_log_var, posteriors_mean, posteriors_log_var, z0

    def postprocess_log_var(self, var: torch.Tensor) -> torch.Tensor:
        """
        Post-processes variance output.

        Args:
            var: Variance

        Returns:
            Post-processes variance output
        """
        if not self._bounded_variance:
            return torch.exp(var)
        else:
            return 0.1 + 0.9 * torch.nn.functional.softplus(var)  # Bounded variance



"""
Implementation of buffered end-to-end filter.
"""
from typing import Tuple

import torch

from motrack_motion.datasets import transforms
from motrack_motion.filter.base import StateModelFilter, State
from motrack_motion.filter.buffer import ObservationBuffer
from motrack_motion.models.architectures.end_to_end import FilterModule


class BufferedE2EFilter(StateModelFilter):
    """
    Implementation of buffered end-to-end
    """
    def __init__(
        self,
        model: FilterModule,
        transform: transforms.InvertibleTransformWithVariance,
        accelerator: str,

        buffer_size: int,
        buffer_min_size: int = 1,
        buffer_min_history: int = 5,
        dtype: torch.dtype = torch.float32,

        recursive_inverse: bool = False
    ):
        """
        Args:
            model: Model
            transform: Transform function
            accelerator: cpu/gpu accelerator
            buffer_size: Buffer size
            buffer_min_size: Minimum size to run model (default 1)
            buffer_min_history: Minimum history to save
            dtype: Torch datatype
            recursive_inverse: Apply recursive inverse transform
        """
        self._model = model
        self._transform = transform
        self._accelerator = accelerator
        self._dtype = dtype

        self._buffer_size = buffer_size
        self._buffer_min_size = buffer_min_size
        self._buffer_min_history = buffer_min_history

        # Model
        self._model.to(self._accelerator)
        self._model.eval()

        self._recursive_inverse = recursive_inverse

    def _get_ts(self) -> Tuple[torch.Tensor, torch.Tensor]:
        t_obs = torch.tensor([0], dtype=self._dtype).view(1, 1)
        t_unobs = torch.tensor([1], dtype=self._dtype).view(1, 1)
        return t_obs, t_unobs

    def initiate(self, measurement: torch.Tensor) -> State:
        buffer = ObservationBuffer(
            size=self._buffer_size,
            min_size=self._buffer_min_size,
            min_history=self._buffer_min_history,
            dtype=self._dtype
        )
        buffer.push(measurement)
        return buffer, None, None, None

    @torch.no_grad()
    def multistep_predict(self, state: State, n_steps: int) -> State:
        assert n_steps == 1, 'Multistep prediction is not supported!'

        buffer, _, _, _ = state
        if not buffer.has_input:
            raise BufferError('Buffer does not have an input!')

        x_obs, ts_obs, ts_unobs = buffer.get_input(n_steps)
        if ts_obs.shape[0] == 1:
            # Only one bbox in history - using baseline (propagate last bbox) instead of NN model
            x_unobs_mean_hat = torch.stack([x_obs[-1].clone() for _ in range(ts_unobs.shape[0])]).to(ts_obs)
            x_unobs_std_hat = 10 * torch.ones_like(x_unobs_mean_hat).to(ts_obs)

            z0, _ = self._model.initialize(batch_size=1, device=self._accelerator)
            t_obs, t_unobs = self._get_ts()
            t_obs, t_unobs = t_obs.to(self._accelerator), t_unobs.to(self._accelerator)
            _, _, z1_prior = self._model.estimate_prior(z0, t_obs, t_unobs)
            z1_prior = z1_prior.detach().cpu()

            return buffer, x_unobs_mean_hat[:, 0, :], x_unobs_std_hat[:, 0, :], z1_prior

        ts_first = ts_obs[0, 0, 0]
        ts_obs = ts_obs - ts_first + 1
        ts_unobs = ts_unobs - ts_first + 1
        t_obs_last, t_unobs_last = ts_obs[-1, 0, 0], ts_unobs[-1, 0, 0]
        if self._recursive_inverse:
            ts_unobs = torch.tensor(list(range(int(t_obs_last) + 1, int(t_unobs_last) + 1)), dtype=torch.float32).view(-1, 1, 1)

        t_x_obs, _, t_ts_obs, t_ts_unobs, *_ = self._transform.apply(data=[x_obs, None, ts_obs, ts_unobs, None], shallow=False)
        t_x_obs, t_ts_obs, t_ts_unobs = t_x_obs.to(self._accelerator), t_ts_obs.to(self._accelerator), t_ts_unobs.to(
            self._accelerator)
        x_unobs_dummy = torch.zeros(1, 1, 4).to(self._accelerator)
        t_x_prior_mean, t_x_prior_var, _, _, z_prior = \
            self._model.forward(t_x_obs, t_ts_obs, x_unobs_dummy, t_ts_unobs, mask=False)
        x_unobs_dummy.detach().cpu()
        t_x_prior_mean = t_x_prior_mean.detach().cpu()
        t_x_prior_mean = t_x_prior_mean.detach().cpu()
        z_prior = z_prior.detach().cpu()

        _, prior_mean, *_ = self._transform.inverse(data=[x_obs, t_x_prior_mean, None], shallow=False)

        prior_mean = prior_mean[-1]  # Removing temporal dimension after inverse transform
        prior_var = self._transform.inverse_var(t_x_prior_var, additional_data=[x_obs, None], shallow=False)
        prior_std = torch.sqrt(prior_var)[0]

        return buffer, prior_mean, prior_std, z_prior

    def predict(self, state: State) -> State:
        buffer, prior_mean, prior_std, z1_prior = self.multistep_predict(state, n_steps=1)
        return buffer, prior_mean[0], prior_std[0], z1_prior

    def singlestep_to_multistep_state(self, state: State) -> State:
        buffer, prior_mean, prior_std, z1_prior = state
        return buffer, prior_mean.unsqueeze(0), prior_std.unsqueeze(0), z1_prior

    @torch.no_grad()
    def update(self, state: State, measurement: torch.Tensor) -> State:
        buffer, _, _, z_prior = state
        x_obs, ts_obs, ts_unobs = buffer.get_input(1)

        if ts_obs.shape[0] == 1:
            # Only one bbox in history - using baseline (propagate last bbox) instead of NN model
            x_unobs_mean_hat = torch.stack([x_obs[-1].clone() for _ in range(ts_unobs.shape[0])]).to(ts_obs)
            x_unobs_std_hat = 10 * torch.ones_like(x_unobs_mean_hat).to(ts_obs)

            z0, _ = self._model.initialize(batch_size=1, device=self._accelerator)
            t_obs, t_unobs = self._get_ts()
            t_obs, t_unobs = t_obs.to(self._accelerator), t_unobs.to(self._accelerator)
            _, _, z1_prior = self._model.estimate_prior(z0, t_obs, t_unobs)
            z1_prior = z1_prior.detach().cpu()

            buffer.push(measurement)
            return buffer, x_unobs_mean_hat[0, 0, :], x_unobs_std_hat[0, 0, :], z1_prior

        _, t_measurement, *_ = self._transform.apply(data=[x_obs, measurement.view(1, 1, -1), ts_obs, ts_unobs, None], shallow=False)
        t_measurement, z_prior = t_measurement[0].to(self._accelerator), z_prior.to(self._accelerator)
        z_evidence = self._model.encode_unobs(t_measurement)
        t_posterior_mean, t_posterior_log_var, z_posterior = self._model.estimate_posterior(z_prior, z_evidence)
        t_posterior_mean, t_posterior_log_var, z_posterior = \
            t_posterior_mean.detach().cpu(), t_posterior_log_var.detach().cpu(), z_posterior.detach().cpu()
        t_posterior_var = self._model.postprocess_log_var(t_posterior_log_var)
        _, posterior_mean, *_ = self._transform.inverse(data=[x_obs, t_posterior_mean.view(1, 1, -1), None], shallow=False)
        posterior_var = self._transform.inverse_var(t_posterior_var, additional_data=[posterior_mean, None], shallow=False)
        posterior_std = torch.sqrt(posterior_var)

        posterior_mean = posterior_mean[0, 0, :]
        posterior_std = posterior_std[0, :]

        buffer.push(measurement)

        return buffer, posterior_mean, posterior_std, z_posterior

    def missing(self, state: State) -> State:
        buffer, mean, std, z_state = state
        buffer.increment()
        return buffer, mean, std, z_state

    def project(self, state: State) -> Tuple[torch.Tensor, torch.Tensor]:
        _, mean, std, _ = state
        var = torch.square(std)
        return mean, var


def run_test() -> None:
    # pylint: disable=import-outside-toplevel
    from motrack_motion.datasets.transforms.base import IdentityTransform
    from motrack_motion.models.architectures.rnn_filter import RNNFilterModel

    smf = BufferedE2EFilter(
        model=RNNFilterModel(
            observable_dim=4,
            latent_dim=4,
            output_dim=4
        ),
        buffer_size=5,
        accelerator='cpu',
        transform=IdentityTransform()
    )

    measurements = torch.randn(10, 4, dtype=torch.float32)

    # Initiate test
    state = smf.initiate(measurements[0])

    # Predict test
    state = smf.predict(state)
    mean_hat, cov_hat = smf.project(state)
    assert mean_hat.shape == (4,) and cov_hat.shape == (4,)

    # Update test
    state = smf.update(state, measurements[1])
    mean_updated, cov_updated = smf.project(state)
    assert mean_updated.shape == (4,) and cov_updated.shape == (4,)


if __name__ == '__main__':
    run_test()



"""
MOT Challenge Dataset support
"""
import configparser
import copy
import enum
import logging
import os
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Tuple, List, Any, Union, Optional

import numpy as np
import pandas as pd
from tqdm import tqdm

from motrack_motion.datasets.common.scene_info import BasicSceneInfo
from motrack_motion.datasets.torch import TrajectoryDataset
from motrack_motion.datasets.utils import split_trajectory_observed_unobserved
from motrack_motion.utils import file_system

CATEGORY = 'pedestrian'
N_IMG_DIGITS = 6

class LabelType(enum.Enum):
    DETECTION = 'det'
    GROUND_TRUTH = 'gt'


@dataclass
class SceneInfo(BasicSceneInfo):
    """
    MOT Scene metadata (name, frame shape, ...)
    """
    dirpath: str
    gt_path: str
    framerate: Union[str, int]
    imdir: str
    imext: str

    def __post_init__(self):
        """
        Convert to proper type
        """
        super().__post_init__()
        self.framerate = int(self.framerate)


SceneInfoIndex = Dict[str, SceneInfo]


logger = logging.getLogger('MOTDataset')


class MOTDataset(TrajectoryDataset):
    """
    Parses MOT dataset in given format
    """
    def __init__(
        self,
        path: str,
        history_len: int,
        future_len: int,
        sequence_list: Optional[List[str]] = None,
        label_type: LabelType = LabelType.GROUND_TRUTH,
        skip_corrupted: bool = False,
        test: bool = False
    ) -> None:
        """
        Args:
            path: Path to dataset
            history_len: Number of observed data points
            future_len: Number of unobserved data points
            sequence_list: Sequence filter by defined list
            label_type: Label Type
            skip_corrupted: Skip corrupted scenes (otherwise error is raised)
            test: Is this dataset used for test evaluation
        """
        super().__init__(
            history_len=history_len,
            future_len=future_len,
            sequence_list=sequence_list
        )

        self._path = path
        self._label_type = label_type

        self._scene_info_index, self._n_digits = self._index_dataset(path, label_type, sequence_list, skip_corrupted, test=test)
        self._data_labels, self._n_labels, self._frame_to_data_index_lookup = self._parse_labels(self._scene_info_index, test=test)
        self._trajectory_index = self._create_trajectory_index(self._data_labels, self._history_len, self._future_len, test=test)

    @property
    def label_type(self) -> LabelType:
        """
        Returns: LabelType
        """
        return self._label_type

    @property
    def scenes(self) -> List[str]:
        return list(self._scene_info_index.keys())

    def parse_object_id(self, object_id: str) -> Tuple[str, str]:
        assert object_id in self._data_labels, f'Unknown object id "{object_id}".'
        scene_name, scene_object_id = object_id.split('_')
        return scene_name, scene_object_id

    def get_object_category(self, object_id: str) -> str:
        return CATEGORY

    def get_scene_object_ids(self, scene_name: str) -> List[str]:
        assert scene_name in self.scenes, f'Unknown scene "{scene_name}". Dataset scenes: {self.scenes}.'
        return [d for d in self._data_labels if d.startswith(scene_name)]

    def get_scene_number_of_object_ids(self, scene_name: str) -> int:
        return len(self.get_scene_object_ids(scene_name))

    def get_object_data_length(self, object_id: str) -> int:
        return len(self._data_labels[object_id])

    def get_object_data_label(self, object_id: str, index: int, relative_bbox_coords: bool = True) -> Optional[dict]:
        data = copy.deepcopy(self._data_labels[object_id][index])
        if data is None:
            return data

        if relative_bbox_coords:
            scene_name, _ = self.parse_object_id(object_id)
            scene_info = self._scene_info_index[scene_name]
            bbox = data['bbox']
            bbox = [
                bbox[0] / scene_info.imwidth,
                bbox[1] / scene_info.imheight,
                bbox[2] / scene_info.imwidth,
                bbox[3] / scene_info.imheight
            ]
            data['bbox'] = bbox

        return data

    def get_object_data_label_by_frame_index(
        self,
        object_id: str,
        frame_index: int,
        relative_bbox_coords: bool = True
    ) -> Optional[dict]:
        return self.get_object_data_label(object_id, frame_index, relative_bbox_coords=relative_bbox_coords)

    def get_scene_info(self, scene_name: str) -> BasicSceneInfo:
        return self._scene_info_index[scene_name]

    def _get_image_path(self, scene_info: SceneInfo, frame_id: int) -> str:
        """
        Get frame path for given scene and frame id

        Args:
            scene_info: scene metadata
            frame_id: frame number

        Returns:
            Path to image (frame)
        """
        return os.path.join(scene_info.dirpath, scene_info.imdir, f'{frame_id:0{self._n_digits}d}{scene_info.imext}')

    def get_scene_image_path(self, scene_name: str, frame_index: int) -> str:
        scene_info = self._scene_info_index[scene_name]
        return self._get_image_path(scene_info, frame_index + 1)

    @staticmethod
    def _get_data_cache_path(path: str, data_name: str) -> str:
        """
        Get cache path for data object path.

        Args:
            path: Path

        Returns:
            Path where data object is or will be stored.
        """
        filename = Path(path).stem
        cache_filename = f'.{filename}.{data_name}.json'
        dirpath = str(Path(path).parent)
        return os.path.join(dirpath, cache_filename)

    @staticmethod
    def parse_scene_ini_file(scene_directory: str, label_type_name) -> SceneInfo:
        gt_path = os.path.join(scene_directory, label_type_name, f'{label_type_name}.txt')
        scene_info_path = os.path.join(scene_directory, 'seqinfo.ini')
        raw_info = configparser.ConfigParser()
        raw_info.read(scene_info_path)
        raw_info = dict(raw_info['Sequence'])
        raw_info['gt_path'] = gt_path
        raw_info['dirpath'] = scene_directory

        return SceneInfo(**raw_info, category=CATEGORY)

    @staticmethod
    def _index_dataset(
        path: str,
        label_type: LabelType,
        sequence_list: Optional[List[str]],
        skip_corrupted: bool,
        test: bool = False
    ) -> Tuple[SceneInfoIndex, int]:
        """
        Index dataset content. Format: { {scene_name}: {scene_labels_path} }

        Args:
            path: Path to dataset
            label_type: Use ground truth bboxes or detections
            sequence_list: Filter scenes
            skip_corrupted: Skips incomplete scenes
            test: Is it test split

        Returns:
            Index to scenes, number of digits used in images name convention (may vary between datasets)
        """
        scene_names = [file for file in file_system.listdir(path) if not file.startswith('.')]
        logger.debug(f'Found {len(scene_names)} scenes. Names: {scene_names}.')
        n_digits = N_IMG_DIGITS

        scene_info_index: SceneInfoIndex = {}
        n_skipped: int = 0

        for scene_name in scene_names:
            if sequence_list is not None and scene_name not in sequence_list:
                continue

            scene_directory = os.path.join(path, scene_name)
            scene_files = file_system.listdir(scene_directory)

            # Scene content validation
            skip_scene = False
            for filename in [label_type.value, 'seqinfo.ini']:
                if filename not in scene_files:
                    if filename == label_type.value and test:
                        continue

                    msg = f'Ground truth file "{filename}" not found on path "{scene_directory}". Contents: {scene_files}'
                    if not skip_corrupted:
                        raise FileNotFoundError(msg)

                    logger.warning(f'Skipping scene "{scene_name}". Reason: `{msg}`')
                    skip_scene = True
                    break

            if 'img1' in scene_files:
                # Check number of digits used in image name (e.g. DanceTrack and MOT20 have different convention)
                img1_path = os.path.join(scene_directory, 'img1')
                image_names = file_system.listdir(img1_path)
                assert len(image_names) > 0, 'Image folder exists but it is empty!'
                image_name = Path(image_names[0]).stem
                n_digits = len(image_name)

            if skip_scene:
                n_skipped += 1
                continue

            scene_info = MOTDataset.parse_scene_ini_file(scene_directory, label_type.value)
            scene_info_index[scene_name] = scene_info
            logger.debug(f'Scene info {scene_info}.')

        if n_digits != N_IMG_DIGITS:
            logger.warning(f'This dataset does not have default number of digits in image name. Got {n_digits} where default is {N_IMG_DIGITS}.')

        logger.info(f'Total number of parsed scenes is {len(scene_info_index)}. Number of skipped scenes is {n_skipped}.')

        return scene_info_index, n_digits

    def _parse_labels(self, scene_infos: SceneInfoIndex, test: bool = False) -> Tuple[Dict[str, List[dict]], int, Dict[str, Dict[int, int]]]:
        """
        Loads all labels dictionary with format:
        {
            {scene_name}_{object_id}: {
                {frame_id}: [ymin, xmin, w, h]
            }
        }

        Args:
            scene_infos: Scene Metadata
            test: If test then no parsing is performed

        Returns:
            Labels dictionary
        """
        data: Dict[str, List[Optional[dict]]] = {}
        frame_to_data_index_lookup: Dict[str, Dict[int, int]] = defaultdict(dict)
        n_labels = 0
        if test:
            # Return empty labels
            return data, n_labels, frame_to_data_index_lookup

        for scene_name, scene_info in scene_infos.items():
            seqlength = self._scene_info_index[scene_name].seqlength

            df = pd.read_csv(scene_info.gt_path, header=None)
            df = df[df[7] == 1]  # Ignoring values that are not evaluated

            df = df.iloc[:, :6]
            df.columns = ['frame_id', 'object_id', 'xmin', 'ymin', 'w', 'h']  # format: yxwh
            df['object_global_id'] = \
                scene_name + '_' + df['object_id'].astype(str)  # object id is not unique over all scenes
            df = df.drop(columns='object_id', axis=1)
            df = df.sort_values(by=['object_global_id', 'frame_id'])
            n_labels += df.shape[0]

            object_groups = df.groupby('object_global_id')
            for object_global_id, df_grp in tqdm(object_groups, desc=f'Parsing {scene_name}', unit='pedestrian'):
                df_grp = df_grp.drop(columns='object_global_id', axis=1).set_index('frame_id')

                data[object_global_id] = [None for _ in range(seqlength)]
                for frame_id, row in df_grp.iterrows():
                    data[object_global_id][int(frame_id) - 1] = {
                        'frame_id': frame_id,
                        'bbox': row.values.tolist(),
                        'image_path': self._get_image_path(scene_info, frame_id),
                        'occ': False,
                        'oov': False,
                        'scene': scene_name,
                        'category': CATEGORY
                    }
                    frame_to_data_index_lookup[object_global_id][frame_id] = len(data[object_global_id]) - 1

        logger.debug(f'Parsed labels. Dataset size is {n_labels}.')
        data = dict(data)  # Disposing unwanted defaultdict side-effects
        return data, n_labels, frame_to_data_index_lookup

    def _create_trajectory_index(
        self,
        labels: Dict[str, List[Optional[dict]]],
        history_len: int,
        future_len: int,
        test: bool = False
    ) -> List[Tuple[str, int, int]]:
        """
        Creates trajectory index by going through every object and creating bbox trajectory of consecutive time points.
        All trajectories that are not continuous are skipped.

        Args:
            labels: List of bboxes for each object
            history_len: Length of observed part of trajectory
            future_len: Length of unobserved part of trajectory
            test: If test then no parsing is performed

        Returns:
            Trajectory index
        """
        trajectory_len = history_len + future_len
        traj_index = []

        if test:
            return traj_index

        # Skipped trajectories stats
        n_skipped = 0
        n_total = 0

        for object_global_id, data in tqdm(labels.items(), desc='Creating trajectories', unit='object'):
            # Fetch scene length
            scene_name, _ = self.parse_object_id(object_global_id)
            seqlength = self._scene_info_index[scene_name].seqlength

            # Add trajectories to the index
            traj_start_time_point_candidates = list(range(seqlength))
            for i in traj_start_time_point_candidates:
                if i + trajectory_len > seqlength:  # Out of range
                    continue

                n_total += 1

                trajectory_range = list(range(i, i + trajectory_len))
                if any((data[frame_id] is None) for frame_id in trajectory_range):
                    # Trajectory is not continuous -> skipping
                    # Note: This is only for motion model training
                    n_skipped += 1
                    continue

                traj_index.append((object_global_id, i, i + trajectory_len))

        n_kept = n_total - n_skipped
        logger.info(f'Total number of trajectory candidates: {n_total}.')
        logger.info(f'Total number of skipped trajectories: {n_skipped} (ratio: {100 * (n_skipped / n_total):.2f}%).')
        logger.info(f'Total number of kept trajectories: {n_kept}.')
        logger.info(f'Trajectory index size: {len(traj_index)}')

        return traj_index

    def __len__(self) -> int:
        return len(self._trajectory_index)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, Dict[str, Any]]:
        object_id, traj_start, traj_end = self._trajectory_index[index]
        raw_traj = self._data_labels[object_id][traj_start:traj_end]
        traj_len = len(raw_traj)
        expected_traj_len = self._history_len + self._future_len
        assert traj_len == self._history_len + self._future_len, f'Expected length {expected_traj_len} but got {traj_len}!'

        scene_name = object_id.split('_')[0]
        scene_info = self._scene_info_index[scene_name]

        # Metadata
        frame_ids = [item['frame_id'] for item in raw_traj]
        metadata = {
            'category': CATEGORY,
            'scene_name': scene_name,
            'object_id': object_id,
            'frame_ids': frame_ids,
            'image_paths': [item['image_path'] for item in raw_traj]
        }

        # Bboxes
        bboxes = np.array([item['bbox'] for item in raw_traj], dtype=np.float32)
        # Normalize bbox coordinates
        bboxes[:, [0, 2]] /= scene_info.imwidth
        bboxes[:, [1, 3]] /= scene_info.imheight

        bboxes_obs, bboxes_unobs, frame_ts_obs, frame_ts_unobs = \
            split_trajectory_observed_unobserved(frame_ids, bboxes, self._history_len)
        return bboxes_obs, bboxes_unobs, frame_ts_obs, frame_ts_unobs, metadata



"""
Training Meters.
"""
from collections import defaultdict
from typing import Tuple, Union, Iterable

import torch


class UnknownMetricException(KeyError):
    """
    No metric value was pushed for this metric name
    """
    pass


class MetricMeter:
    """
    Metric/Loss Meter for training
    """
    def __init__(self):
        self._history = defaultdict(list)

    def push(self, name: str, value: Union[torch.Tensor, float]) -> None:
        """
        Pushes new metric value. List of these values are later used to calculate average metric value
        by using `get()` or `get_all()` method.

        Args:
            name: Metric name
            value: Metric value
        """
        self._history[name].append(value)

    def get(self, name: str) -> Union[torch.Tensor, float]:
        """
        Calculates average over all pushed metric values.

        Args:
            name: Name of the metric

        Returns:
            Metric score
        """
        if name not in self._history:
            raise UnknownMetricException(f'Metric name "{name}" not found. '
                                         f'Known metrics: {list(self._history.keys())}.')

        values = self._history[name]
        return sum(values) / len(values)

    def get_all(self, flush: bool = True) -> Iterable[Tuple[str, Union[torch.Tensor, float]]]:
        """
        Calculates averages over all pushed metric values for every pushed metrics.
        Optionally deletes all data if flush is activated (default).

        Args:
            flush: Performs flush at the end

        Returns:
            Iterator: name of the metric, metric score
        """
        for name in self._history:
            yield name, self.get(name)

        if flush:
            self.flush()

    def flush(self) -> None:
        """
        Deletes all pushed data.
        """
        self._history = defaultdict(list)



"""
Performs original "train" dataset split into halves where
the first half belongs to the train
and the second half belongs to the validation.

Steps:
1. Parse seqinfo file and extract seqlength
2. Split gt/gt.txt and det/det.txt (csv files)
3. Copy images (or create symlinks)
4. Create new seqinfo files
"""
import argparse
import configparser
import logging
import os
from pathlib import Path

import pandas as pd
from tqdm import tqdm

from motrack_motion.utils import file_system
from motrack_motion.utils.logging import configure_logging

logger = logging.getLogger('MOT20Split')


def parse_args() -> argparse.Namespace:
    """
    Scene split parse configuration.

    Example run:
        python3 motrack_motion/datasets/mot/tools/run_split_train_into_two_halves.py \
            --input-path {ASSETS_PATH}/MOT17/train \
            --output-path {ASSETS_PATH}/MOT17

    Returns:
        Scene split configuration
    """
    parser = argparse.ArgumentParser(description='Split MOT configuration')
    parser.add_argument('--input-path', type=str, required=True, help='MOT dataset path.')
    parser.add_argument('--output-path', type=str, required=True, help='MOT new dataset path')
    return parser.parse_args()


def parse_ini(path: str, subject: str) -> dict:
    """
    Parse an .ini file and return a dictionary of the specified section.

    Args:
        path: The file path of the .ini file to be read.
        subject: The section in the .ini file to be converted to a dictionary.

    Returns:
        A dictionary containing the key-value pairs of the specified section
        in the .ini file.
    """
    raw_info = configparser.ConfigParser()
    raw_info.read(path)
    return dict(raw_info[subject])


def save_ini(data: dict, path: str, subject: str) -> None:
    """
    Save a dictionary into an .ini file.

    Args:
        data: The dictionary to be saved.
        path: The file path of the .ini file.
        subject: The header under which the dictionary will be saved.
    """
    config = configparser.ConfigParser()
    config[subject] = data

    with open(path, 'w', encoding='utf-8') as configfile:
        config.write(configfile)


def split_scene(input_path: str, output_path: str, scene: str) -> None:
    """
    Split a MOT scene into two halves and create new folders containing the respective information.

    Args:
        input_path: The directory containing the MOT scene.
        output_path: The directory where the two new scenes will be created.
        scene: The name of the scene.
    """
    scene_path = os.path.join(input_path, scene)
    seqinfo = parse_ini(os.path.join(scene_path, 'seqinfo.ini'), 'Sequence')
    seqlength = int(seqinfo['seqlength'])
    middle_frame = seqlength // 2

    # Create new directories for the two halves
    first_scene_name = scene + '-H1'
    second_scene_name = scene + '-H2'
    first_half_path = os.path.join(output_path, 'train', first_scene_name)
    second_half_path = os.path.join(output_path, 'val', second_scene_name)

    Path(first_half_path).mkdir(parents=True, exist_ok=True)
    Path(second_half_path).mkdir(parents=True, exist_ok=True)

    # Split gt.txt and det.txt using pandas
    for file in ['gt/gt.txt', 'det/det.txt']:
        df = pd.read_csv(os.path.join(scene_path, file), header=None)
        first_half = df[df[0] <= middle_frame]
        second_half = df[df[0] > middle_frame].copy()
        second_half[0] = second_half[0] - middle_frame

        first_half_filepath = os.path.join(first_half_path, file)
        Path(first_half_filepath).parent.mkdir(exist_ok=True, parents=True)
        first_half.to_csv(first_half_filepath, header=None, index=False)

        second_half_filepath = os.path.join(second_half_path, file)
        Path(second_half_filepath).parent.mkdir(exist_ok=True, parents=True)
        second_half.to_csv(second_half_filepath, header=None, index=False)

    # Write new seqinfo.ini files for the two halves
    seqinfo.pop('seqlength')
    seqinfo.pop('name')
    first_half_seqinfo = {**seqinfo, 'seqlength': str(middle_frame), 'name': first_scene_name}
    second_half_seqinfo = {**seqinfo, 'seqlength': str(seqlength - middle_frame), 'name': second_scene_name}

    save_ini(first_half_seqinfo, os.path.join(first_half_path, 'seqinfo.ini'), 'Sequence')
    save_ini(second_half_seqinfo, os.path.join(second_half_path, 'seqinfo.ini'), 'Sequence')

    # Create symlinks for images
    img_path = os.path.join(scene_path, 'img1')
    for i in range(1, seqlength + 1):
        source_img = os.path.join(img_path, f'{i:06}.jpg')
        dst_img = os.path.join(first_half_path, 'img1', f'{i:06}.jpg') if i <= middle_frame \
            else os.path.join(second_half_path, 'img1', f'{i - middle_frame:06}.jpg')
        Path(dst_img).parent.mkdir(exist_ok=True, parents=True)
        os.symlink(source_img, dst_img)


def main(args: argparse.Namespace) -> None:
    input_path: str = args.input_path
    output_path: str = args.output_path

    scenes = file_system.listdir(input_path)
    n_scenes = len(scenes)
    logger.info(f'Indexed {n_scenes} scenes: {scenes}.')

    for scene in tqdm(scenes, total=len(scenes), unit='scene'):
        split_scene(
            scene=scene,
            input_path=input_path,
            output_path=output_path
        )


if __name__ == '__main__':
    configure_logging()
    main(parse_args())



"""
Loss function factory method.
"""
from torch import nn

from motrack_motion.models.losses.giou import HybridGaussianNLLLossGIoU, HybridL1GIoU


def factory_loss_function(name: str, params: dict) -> nn.Module:
    """
    Crates loss function based on give name and parameters.

    Args:
        name: Loss function name
        params: Loss function parameters

    Returns:
        Loss function object.
    """
    catalog = {
        'mse': nn.MSELoss,
        'gaussian_nllloss': nn.GaussianNLLLoss,
        'hybrid_l1_giou': HybridL1GIoU,
        'hybrid_gaussian_nllloss_giou': HybridGaussianNLLLossGIoU
    }

    name = name.lower()
    if name not in catalog:
        raise KeyError(f'Unknown loss name "{name}". Available: {list(catalog.keys())}.')

    return catalog[name](**params)



"""
End-to-end filter training module.
"""
import random
from typing import Optional, Union, List, Tuple, Dict

import torch

from motrack_motion.datasets.augmentations.trajectory import remove_points
from motrack_motion.datasets.transforms import InvertibleTransform, InvertibleTransformWithVariance
from motrack_motion.models.architectures.end_to_end import FilterModule
from motrack_motion.models.losses import factory_loss_function
from motrack_motion.models.train.base import LightningModuleBase, LightningTrainConfig
from motrack_motion.models.train.metrics import metrics_func


class EndToEndFilterTrainingModule(LightningModuleBase):
    """
    PytorchLightning wrapper for FilterModule model.
    """
    def __init__(
        self,
        model: FilterModule,

        transform_func: Optional[Union[InvertibleTransform, InvertibleTransformWithVariance]] = None,

        augmentation_remove_points_enable: bool = False,
        augmentation_remove_random_points_proba: float = 0.15,
        augmentation_remove_sequence_proba: float = 0.15,

        train_config: Optional[LightningTrainConfig] = None,
        log_epoch_metrics: bool = True
    ):
        super().__init__(train_config=train_config)

        assert isinstance(model, FilterModule), f'Expected FilterModule but found {type(model)}'
        self._model = model

        self._augmentation_remove_points_enable = augmentation_remove_points_enable
        self._augmentation_remove_random_points_proba = augmentation_remove_random_points_proba
        self._augmentation_remove_sequence_proba = augmentation_remove_sequence_proba

        self._loss_func = factory_loss_function(train_config.loss_name, train_config.loss_params) \
            if train_config is not None else None
        if train_config is not None:
            assert 'gaussian_nllloss' in train_config.loss_name, \
                'Failed to find "gaussian_nllloss" in loss function name!'
        if transform_func is not None:
            assert isinstance(transform_func, InvertibleTransformWithVariance), \
                f'Expected transform function to be of type "InvertibleTransformWithStd" ' \
                f'but got "{type(transform_func)}"'
        self._transform_func = transform_func
        self._log_epoch_metrics = log_epoch_metrics

    @property
    def core(self) -> FilterModule:
        """
        Get core model (without training module).
        Returns:
            Core model
        """
        return self._model

    def forward(self, x_obs: torch.Tensor, t_obs: torch.Tensor, x_unobs: torch.Tensor, t_unobs: torch.Tensor,
                mask: Union[bool, List[bool]] = True, *args, **kwargs) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        return self._model(x_obs, t_obs, x_unobs, t_unobs, mask=mask, *args, **kwargs)

    def inference(self, x_obs: torch.Tensor, t_obs: torch.Tensor, x_unobs: torch.Tensor, t_unobs: torch.Tensor,
                  mask: Union[bool, List[bool]] = True, *args, **kwargs) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Inference (alias for forward)

        Args:
            x_obs: Observed data
            t_obs: Observed time points
            x_unobs: Unobserved data
            t_unobs: Unobserved time points
            mask: Mask unobserved

        Returns:
            Prior and posterior estimation for future trajectory
        """
        return self._model(x_obs, t_obs, x_unobs, t_unobs, mask=mask, *args, **kwargs)

    def _calc_loss_and_metrics(
        self,
        orig_bboxes_obs: torch.Tensor,
        orig_bboxes_unobs_prior: torch.Tensor,
        orig_bboxes_unobs_posterior: torch.Tensor,
        transformed_bboxes_unobs_prior: torch.Tensor,
        transformed_bboxes_unobs_posterior: torch.Tensor,
        bboxes_unobs_prior_mean: torch.Tensor,
        bboxes_unobs_prior_log_var: torch.Tensor,
        bboxes_unobs_posterior_mean: torch.Tensor,
        bboxes_unobs_posterior_log_var: torch.Tensor,
        metadata: dict
    ) -> Tuple[Union[torch.Tensor, Dict[str, torch.Tensor]], Optional[Dict[str, float]]]:
        bboxes_unobs_prior_var = self._model.postprocess_log_var(bboxes_unobs_prior_log_var)
        bboxes_unobs_posterior_var = self._model.postprocess_log_var(bboxes_unobs_posterior_log_var)

        prior_loss = self._loss_func(bboxes_unobs_prior_mean, transformed_bboxes_unobs_prior, bboxes_unobs_prior_var)
        posterior_loss = self._loss_func(bboxes_unobs_posterior_mean, transformed_bboxes_unobs_posterior, bboxes_unobs_posterior_var)
        loss = {
            'prior_loss': prior_loss,
            'posterior_loss': posterior_loss,
            'loss': prior_loss + posterior_loss
        }

        if self._transform_func is not None:
            # Invert mean
            _, bboxes_unobs_prior_mean, *_ = self._transform_func.inverse([orig_bboxes_obs, bboxes_unobs_prior_mean, metadata, None], shallow=False)
            _, bboxes_unobs_posterior_mean, *_ = self._transform_func.inverse([orig_bboxes_obs, bboxes_unobs_posterior_mean, metadata, None],
                                                                              shallow=False)

        if not self._log_epoch_metrics:
            return loss, None

        gt_traj = orig_bboxes_unobs_prior.detach().cpu().numpy()
        prior_traj = bboxes_unobs_prior_mean.detach().cpu().numpy()
        prior_metrics = metrics_func(gt_traj, prior_traj)
        prior_metrics = {f'prior_{name}': value for name, value in prior_metrics.items()}

        gt_traj = orig_bboxes_unobs_posterior.detach().cpu().numpy()
        posterior_traj = bboxes_unobs_posterior_mean.detach().cpu().numpy()
        posterior_metrics = metrics_func(gt_traj, posterior_traj) if self._log_epoch_metrics else None
        posterior_metrics = {f'posterior_{name}': value for name, value in posterior_metrics.items()}
        metrics = dict(list(prior_metrics.items()) + list(posterior_metrics.items()))

        return loss, metrics


    def _log_loss(self, loss: Union[torch.Tensor, Dict[str, torch.Tensor]], prefix: str, log_step: bool = True) -> None:
        """
        Helper function to log loss. Options:
        - Single value: logged as "{prefix}/loss"
        - Dictionary: for each key log value as "{prefix}/{key}"

        Args:
            loss: Loss
            prefix: Prefix (train or val)
        """
        assert prefix in ['training', 'val'], f'Invalid prefix value "{prefix}"!'

        if isinstance(loss, dict):
            assert 'loss' in loss, \
                f'When returning loss as dictionary it has to have key "loss". Found: {list(loss.keys())}'
            for name, value in loss.items():
                value = value.detach().cpu()
                assert not torch.isnan(value).any(), f'Got nan value for key "{name}"!'
                self._meter.push(f'{prefix}-epoch/{name}', value)
                if log_step:
                    self.log(f'{prefix}/{name}', value, prog_bar=False)
        else:
            loss = loss.detach().cpu()
            assert not torch.isnan(loss).any(), 'Got nan value!'
            loss = loss.detach().cpu()
            self._meter.push(f'{prefix}-epoch/loss', loss)
            if log_step:
                self.log(f'{prefix}/loss', loss, prog_bar=False)

    def _log_metrics(self, metrics: Optional[Dict[str, float]], prefix: str) -> None:
        """
        Helper function to log metrics. Input format:
        - Dictionary: for each key log value as "{prefix}-epoch/{key}"

        Args:
            prefix: Prefix (train or val)
        """
        if metrics is None:
            return

        assert prefix in ['training', 'val'], f'Invalid prefix value "{prefix}"!'
        for name, value in metrics.items():
            self._meter.push(f'{prefix}-metrics/{name}', value)

    def training_step(self, batch: Dict[str, Union[dict, torch.Tensor]], *args, **kwargs) -> torch.Tensor:
        bboxes_obs, bboxes_aug_unobs, ts_obs, ts_unobs, orig_bboxes_obs, orig_bboxes_unobs, bboxes_unobs, metadata = batch.values()

        bboxes_unobs_posterior = bboxes_unobs
        orig_bboxes_unobs_posterior = orig_bboxes_unobs
        mask = [True for _ in range(ts_unobs.shape[0])]
        if self._augmentation_remove_points_enable:
            with torch.no_grad():
                assert bboxes_aug_unobs.shape[0] >= 3, 'Minimum length of trajectory is 3 in order to perform special augmentations!'

                r = random.random()
                if r < self._augmentation_remove_random_points_proba:
                    (bboxes_aug_unobs,), removed_indices, kept_indices = remove_points([bboxes_aug_unobs], min_length=1)
                    for i in removed_indices:
                        mask[i] = False
                    bboxes_unobs_posterior = bboxes_unobs_posterior[kept_indices, :, :]
                    orig_bboxes_unobs_posterior = orig_bboxes_unobs_posterior[kept_indices, :, :]

                elif r < self._augmentation_remove_random_points_proba + self._augmentation_remove_sequence_proba:
                    # Remove random sequence (connected points)
                    n = bboxes_aug_unobs.shape[0]
                    start = random.randint(1, n - 1)  # s ~ U[1, n-1]
                    end = random.randint(start, n)  # e ~ U[s, n]
                    bboxes_aug_unobs = torch.cat([bboxes_aug_unobs[:start, :, :], bboxes_aug_unobs[end:n, :, :]], dim=0)
                    for i in range(start, end):
                        mask[i] = False
                    bboxes_unobs_posterior = torch.cat([bboxes_unobs_posterior[:start, :, :], bboxes_unobs_posterior[end:n, :, :]], dim=0)
                    orig_bboxes_unobs_posterior = \
                        torch.cat([orig_bboxes_unobs_posterior[:start, :, :], orig_bboxes_unobs_posterior[end:n, :, :]], dim=0)

        bboxes_prior_mean, bboxes_prior_log_var, bboxes_posterior_mean, bboxes_posterior_log_var, _ = \
            self.forward(bboxes_obs, ts_obs, bboxes_aug_unobs, ts_unobs, mask=mask)

        # noinspection PyTypeChecker
        loss, metrics = self._calc_loss_and_metrics(
            orig_bboxes_obs=orig_bboxes_obs,
            orig_bboxes_unobs_prior=orig_bboxes_unobs,
            orig_bboxes_unobs_posterior=orig_bboxes_unobs_posterior,
            transformed_bboxes_unobs_prior=bboxes_unobs,
            transformed_bboxes_unobs_posterior=bboxes_unobs_posterior,
            bboxes_unobs_prior_mean=bboxes_prior_mean,
            bboxes_unobs_prior_log_var=bboxes_prior_log_var,
            bboxes_unobs_posterior_mean=bboxes_posterior_mean,
            bboxes_unobs_posterior_log_var=bboxes_posterior_log_var,
            metadata=metadata
        )
        self._log_loss(loss, prefix='training', log_step=True)
        self._log_metrics(metrics, prefix='training')

        return loss

    def validation_step(self, batch: Dict[str, Union[dict, torch.Tensor]], *args, **kwargs) -> torch.Tensor:
        bboxes_obs, bboxes_aug_unobs, ts_obs, ts_unobs, orig_bboxes_obs, orig_bboxes_unobs, bboxes_unobs, metadata = batch.values()

        bboxes_prior_mean, bboxes_prior_log_var, bboxes_posterior_mean, bboxes_posterior_log_var, _ = \
            self.forward(bboxes_obs, ts_obs, bboxes_aug_unobs, ts_unobs)

        # noinspection PyTypeChecker
        loss, metrics = self._calc_loss_and_metrics(
            orig_bboxes_obs=orig_bboxes_obs,
            orig_bboxes_unobs_prior=orig_bboxes_unobs,
            orig_bboxes_unobs_posterior=orig_bboxes_unobs,
            transformed_bboxes_unobs_prior=bboxes_unobs,
            transformed_bboxes_unobs_posterior=bboxes_unobs,
            bboxes_unobs_prior_mean=bboxes_prior_mean,
            bboxes_unobs_prior_log_var=bboxes_prior_log_var,
            bboxes_unobs_posterior_mean=bboxes_posterior_mean,
            bboxes_unobs_posterior_log_var=bboxes_posterior_log_var,
            metadata=metadata
        )
        self._log_loss(loss, prefix='val', log_step=False)
        self._log_metrics(metrics, prefix='val')

        return loss



"""
Set of time-series transformation functions.
"""
import torch


def first_difference(xs: torch.Tensor) -> torch.Tensor:
    """
    Transforms time series to first difference. Let X be the data stream and Y the result:
        Y[i] = X[i] - X[i-1] if i > 0
        Y[0] = 0

    Args:
        xs: Time-series (data stream)

    Returns:
        First difference of the time-series
    """
    diff = torch.zeros_like(xs)
    diff[0] = 0
    diff[1:] = xs[1:] - xs[:-1]
    diff = diff.to(xs)
    return diff



"""
Implementation of trajectory augmentations.
"""
import random
from abc import abstractmethod, ABC
from typing import Tuple, List

import torch


class TrajectoryAugmentation(ABC):
    """
    Abstract augmentation - defines interface
    """
    @abstractmethod
    def apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Args:
            x_obs: Observed bboxes data
            x_unobs: Unobserved bboxes data
            t_obs: Observed time points
            t_unobs: Unobserved time points

        Returns:
            augmented (x_obs, x_unobs, t_obs, t_unobs)
        """
        pass

    def __call__(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Alias for `apply` function.

        Args:
            x_obs: Observed bboxes data
            x_unobs: Unobserved bboxes data
            t_obs: Observed time points
            t_unobs: Unobserved time points

        Returns:
            augmented (x_obs, t_obs, x_unobs, t_unobs)
        """
        return self.apply(x_obs, x_unobs, t_obs, t_unobs)


class IdentityAugmentation(TrajectoryAugmentation):
    """
    Performs no transformations (identity).
    """
    def apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        return x_obs, x_unobs, t_obs, t_unobs


class CompositionAugmentation(TrajectoryAugmentation):
    """
    Composition of multiple augmentations.
    """
    def __init__(self, augs: List[TrajectoryAugmentation]):
        """
        Args:
            augs: List of augmentations
        """
        self._augs = augs

    def apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        for aug in self._augs:
            x_obs, x_unobs, t_obs, t_unobs = aug.apply(x_obs, x_unobs, t_obs, t_unobs)
        return x_obs, x_unobs, t_obs, t_unobs



class NonDeterministicAugmentation(TrajectoryAugmentation, ABC):
    """
    Non-deterministic augmentation.
    """
    def __init__(self, proba: float):
        """
        Args:
            proba: Probability to apply augmentation.
        """
        self._proba = proba

    def apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        r = random.uniform(0, 1)
        if r > self._proba:
            # Skip augmentation
            return x_obs, x_unobs, t_obs, t_unobs

        return self._apply(x_obs, x_unobs, t_obs, t_unobs)

    @abstractmethod
    def _apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Args:
            x_obs: Observed bboxes data
            x_unobs: Unobserved bboxes data
            t_obs: Observed time points
            t_unobs: Unobserved time points

        Returns:
            augmented (x_obs, x_unobs, t_obs, t_unobs)
        """


class GaussianNoiseAugmentation(NonDeterministicAugmentation):
    """
    Add Gaussian noise based on the bbox width and height.
    """
    def __init__(self, sigma: float = 0.05, proba: float = 0.5, unobs_noise: bool = False):
        """
        Args:
            sigma: Noise multiplier
            proba: Probability to apply this augmentation
            unobs_noise: Apply noise to unobserved part of trajectory
        """
        super().__init__(proba=proba)
        self._sigma = sigma
        self._unobs_noise = unobs_noise

    def _add_noise(self, x: torch.Tensor) -> torch.Tensor:
        """
        Adds guassian vector to bboxes vector x.

        Args:
            x:

        Returns:
            Input vector with noise
        """
        x_noise = self._sigma * torch.randn_like(x)
        x_noise[..., 0] *= x[..., 2]  # `x` noise is proportional to the `h`
        x_noise[..., 2] *= x[..., 2]  # `h` noise is proportional to the `h`
        x_noise[..., 1] *= x[..., 3]  # `y` noise is proportional to the `w`
        x_noise[..., 3] *= x[..., 3]  # `w` noise is proportional to the `w`
        return x + x_noise


    def _apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        x_obs = self._add_noise(x_obs)
        if self._unobs_noise:
            x_unobs = self._add_noise(x_unobs)

        return x_obs, x_unobs, t_obs, t_unobs


class ShortenTrajectoryAugmentation(NonDeterministicAugmentation):
    """
    Shortens the input trajectory.
    """
    def __init__(self, min_length: int, proba: float):
        """
        Args:
            min_length: Min Trajectory length
                - Augmented trajectory can't be shortened
                  if it already has length `min_length` or less
            proba: Probability to apply
        """
        super().__init__(proba=proba)
        self._min_length = min_length

    def _apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        traj_length = x_obs.shape[0]
        if traj_length <= self._min_length:
            return x_obs, x_unobs, t_obs, t_unobs

        new_traj_length = round(random.uniform(self._min_length, traj_length))
        traj_start = traj_length - new_traj_length
        x_obs = x_obs[traj_start:, ...]
        t_obs = t_obs[traj_start:, ...]

        return x_obs, x_unobs, t_obs, t_unobs


def remove_points(xs: List[torch.Tensor], min_length: int) -> Tuple[List[torch.Tensor], List[int], List[int]]:
    """
    Helper function that removes random number of points from trajectory.

    Args:
        xs: Trajectory components (e.g. values, time, ...)
            - Same operation is performed on each of these tensors
            - It is assumed that their first dimension is same and that they are 3D
        min_length: Minimum result trajectory length

    Returns:
        Trajectory with some removed points, List of removed points, List of kept points
    """
    n_points = xs[0].shape[0]
    max_points_to_remove = max(0, n_points - min_length)
    n_points_to_remove = random.randrange(0, max_points_to_remove)
    all_point_indices = list(range(n_points))
    points_to_remove = random.sample(all_point_indices, k=n_points_to_remove)
    points_to_keep = [point for point in all_point_indices if point not in points_to_remove]

    return [x[points_to_keep, :, :] for x in xs], points_to_remove, points_to_keep


class RemoveRandomPointsTrajectoryAugmentation(NonDeterministicAugmentation):
    """
    Remove random points from input trajectory.
    """
    def __init__(self, min_length: int, proba: float):
        """
        Args:
            min_length: Min Trajectory length
                - Augmented trajectory can't be shortened more than `min_length`
            proba: Probability to apply
        """
        super().__init__(proba=proba)
        self._min_length = min_length

    def _apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        (x_obs, t_obs), _, _ = remove_points([x_obs, t_obs], self._min_length)
        return x_obs, x_unobs, t_obs, t_unobs


class RemoveRandomPointsUnobservedTrajectoryAugmentation(NonDeterministicAugmentation):
    """
    Remove random points from target trajectory. Same as `RemoveRandomPointsTrajectoryAugmentation` but for unobserved trajectory.
    """
    def __init__(self, min_length: int, proba: float):
        """
        Args:
            min_length: Min Trajectory length
                - Augmented trajectory can't be shortened more than `min_length`
            proba: Probability to apply
        """
        super().__init__(proba=proba)
        self._min_length = min_length
        self._proba = proba

    def _apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        (x_unobs, t_unobs), _, _ = remove_points([x_unobs, t_unobs], self._min_length)
        return x_obs, x_unobs, t_obs, t_unobs


class CameraMovementAugmentation(NonDeterministicAugmentation):
    """
    Simulates linear camera movement.
    """
    def __init__(self, max_magnitude: float, proba: float):
        """
        Args:
            max_magnitude: Maximum camera movement magnitude
            proba: Probability to apply the augmentation
        """
        super().__init__(proba=proba)
        self._max_magnitude = max_magnitude

    def _apply(self, x_obs: torch.Tensor, x_unobs: torch.Tensor, t_obs: torch.Tensor, t_unobs: torch.Tensor) \
        -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:

        angle = torch.tensor(random.uniform(-0.95, 0.95) * torch.pi, dtype=torch.float32)
        m = torch.tensor(random.uniform(-1, 1) * self._max_magnitude)
        n_obs_steps, n_unobs_steps = t_obs.shape[0], t_unobs.shape[0]
        n_steps = n_obs_steps + n_unobs_steps
        x_end, y_end = m * torch.cos(angle), m * torch.sin(angle)

        obs_steps = torch.tensor(list(range(n_obs_steps)), dtype=torch.float32) / (n_steps - 1)
        unobs_steps = torch.tensor(list(range(n_obs_steps, n_obs_steps + n_unobs_steps)), dtype=torch.float32) / (n_steps - 1)
        x_obs[..., 0] += x_end * obs_steps.expand_as(x_obs[..., 0])
        x_obs[..., 1] += y_end * obs_steps.expand_as(x_obs[..., 1])
        x_unobs[..., 0] += x_end * unobs_steps.expand_as(x_unobs[..., 0])
        x_unobs[..., 1] += y_end * unobs_steps.expand_as(x_unobs[..., 1])
        return x_obs, x_unobs, t_obs, t_unobs


def create_identity_augmentation_config() -> dict:
    """
    Returns:
        Identity augmentation (no transformation) config
    """
    return {
        '_target_': 'motrack_motion.datasets.augmentations.trajectory.IdentityAugmentation',
    }



"""
Logging Utility
"""
import logging

import numpy as np
import torch

logger = logging.getLogger('UtilsLogging')


def configure_logging(level: int = logging.INFO) -> None:
    """
    Configures logger

    Args:
        level: Logging level
    """
    np.set_printoptions(formatter={'float': '{:0.3f}'.format})
    torch.set_printoptions(sci_mode=False, precision=3)
    logging.basicConfig(
        level=logging.getLevelName(level),
        format='%(asctime)s [%(name)s]: %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    logging.getLogger('matplotlib').setLevel(logging.CRITICAL)



"""
Filter's observation buffer.
"""
from typing import Tuple, List

import torch


class ObservationBuffer:
    """
    Buffers trajectory observations.
    """
    def __init__(
        self,
        size: int,
        min_size: int,
        min_history: int = 1,
        dtype: torch.dtype = torch.float32,
    ):
        assert size >= 1, f'Invalid size {size}. Minimum size is 1.'

        self._size = size
        self._min_size = min_size
        self._dtype = dtype
        self._min_history = min_history

        self._buffer: List[Tuple[int, torch.Tensor]] = []
        self._t = 0

    @property
    def time(self) -> int:
        return self._t

    @property
    def has_input(self) -> bool:
        return len(self._buffer) >= self._min_size

    def push(self, x: torch.Tensor) -> None:
        self._buffer.append((self._t, x))
        self.increment()

    def increment(self) -> None:
        if len(self._buffer) > self._min_history and (self._t - self._buffer[0][0]) >= self._size:
            self._buffer.pop(0)

        self._t += 1

    def get_input(self, n_future_steps: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        n_hist_steps = len(self._buffer)
        ts_obs, x_obs = zip(*self._buffer)
        ts_obs_first = ts_obs[0]

        # Form history trajectory
        x_obs = torch.stack(x_obs).view(n_hist_steps, 1, -1)
        ts_obs = torch.tensor(ts_obs, dtype=self._dtype).view(-1, 1, 1)

        # Form estimation trajectory time interval
        ts_unobs = torch.tensor(list(range(self._t, self._t + n_future_steps)),
                                dtype=self._dtype).view(-1, 1, 1)

        ts_obs = ts_obs - ts_obs_first + 1
        ts_unobs = ts_unobs - ts_obs_first + 1

        return x_obs, ts_obs, ts_unobs

    def clear(self) -> None:
        self._buffer.clear()

    def __len__(self) -> int:
        return len(self._buffer)



"""
Model factory method
"""
import enum
from typing import Union

from torch import nn

from motrack_motion.models.architectures.rnn_filter import RNNFilterModel


class ModelType(enum.Enum):
    """
    Enumerated implemented architectures
    """
    # RNN
    RNN_FILTER = 'rnn_filter'

    @classmethod
    def from_str(cls, value: str) -> 'ModelType':
        for v in cls:
            if v.value.lower() == value.lower():
                return v

        raise ValueError(f'Can\'t create ModelType from "{value}". Possible values: {list(cls)}')


def model_factory(
    model_type: Union[ModelType, str],
    params: dict,
) -> nn.Module:
    """
    Model factory method.

    Args:
        model_type: Model type
        params: Model parameters'

    Returns:
        Model
    """
    if isinstance(model_type, str):
        model_type = ModelType.from_str(model_type)

    catalog = {
        ModelType.RNN_FILTER: RNNFilterModel
    }

    return catalog[model_type](**params)




"""
Torch dataset support. Any Dataset that implements `TrajectoryDataset` interface can be used for training and evaluation.
"""
from motrack_motion.datasets.torch.core import TorchTrajectoryDataset, TrajectoryDataset



"""
Standard models for trajectory prediction
"""
from motrack_motion.models.architectures.factory import model_factory
from motrack_motion.models.architectures.rnn_filter import RNNFilterModel
from motrack_motion.models.train.end_to_end import EndToEndFilterTrainingModule
from motrack_motion.models.train.factory import load_or_create_training_module



"""
BBox transformations. Contains
- BBox trajectory first order difference transformation
- BBox coordination standardization to N(0, 1)
- BBox trajectory standardized first order difference transformation
- BBox trajectory relative to the last observed point
- BBox trajectory standardized relative to the last observed point transformation
- BBox transform composition
- BBox coordination standardization to N(0, 1) (by category)
"""

from typing import Union, List, Optional

import torch

from motrack_motion.datasets.transforms.base import InvertibleTransformWithVariance, TensorCollection


class BboxFirstOrderDifferenceTransform(InvertibleTransformWithVariance):
    """
    Applies first difference transformation:
    Y[i] = X[i] - X[i-1]
    Y[0] is removed

    If `scale_by_time_diff=True` then:
    Y[i] = X[i] - X[i-1] / (T[i] - T[i-1])
    """
    def __init__(self, scale_by_time_diff: bool = False):
        """
        Args:
            scale_by_time_diff: Scale difference by time difference
                - Useful in case of missing observed points
        """
        super().__init__(name='first_difference')
        self._scale_by_time_diff = scale_by_time_diff

    def apply(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        bbox_obs, bbox_unobs, ts_obs, ts_unobs, *other = data
        assert bbox_obs.shape[0] >= 0, f'{self.__name__} requires at least 2 observable points. ' \
                                       f'Found {bbox_obs.shape[0]}'

        assert not shallow
        if not shallow:
            bbox_obs = bbox_obs.clone()
            bbox_unobs = bbox_unobs.clone() if bbox_unobs is not None else None
            ts_obs = ts_obs.clone()

        if bbox_unobs is not None:
            # During live inference, bbox_unobs are not known (this is for training only)
            bbox_unobs[1:, ...] = bbox_unobs[1:, ...] - bbox_unobs[:-1, ...]
            bbox_unobs[0, ...] = bbox_unobs[0, ...] - bbox_obs[-1, ...]

        bbox_obs[1:, ...] = bbox_obs[1:, ...] - bbox_obs[:-1, ...]
        if self._scale_by_time_diff:
            bbox_obs[1:, ...] = bbox_obs[1:, ...] / (ts_obs[1:, ...] - ts_obs[:-1, ...])

        bbox_obs, ts_obs = bbox_obs[1:, ...], ts_obs[1:, ...]  # Dump first

        return bbox_obs, bbox_unobs, ts_obs, ts_unobs, *other

    def inverse(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        orig_bbox_obs, bbox_hat, *other = data
        if not shallow:
            bbox_hat = bbox_hat.clone()

        bbox_hat[0, ...] = bbox_hat[0, ...] + orig_bbox_obs[-1, ...]
        bbox_hat = torch.cumsum(bbox_hat, dim=0)

        return orig_bbox_obs, bbox_hat, *other

    def inverse_std(
        self,
        t_std: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        # Check `inverse_var`
        t_var = torch.square(t_std)
        t_var_cumsum = torch.cumsum(t_var, dim=0)
        t_std_cumsum = torch.sqrt(t_var_cumsum)
        return t_std_cumsum

    def inverse_var(
        self,
        t_var: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        # In order to estimate std for inverse two assumptions are added
        # 1. Random variables y[i-1] and y[i] are independent
        # 2. Variance of last observed bbox coordinate is 0 (i.e. var(x[-1]) = 0)
        # Let: y - transformed, x - original, 0 - first unobserved time point
        # Transformation: y[i] = x[i] - x[i-1]
        # Inverse transformation: x[i] = y[i] + x[i-1]
        # Known variances: y[i] and x[-1]
        # => var(x[0]) = var(y[0]) + var(x[-1]) = var(y[0])  # from (1) and (2)
        # => var(x[1]) = var(y[1]) + var(x[0]) = var(y[0]) + var(y[1]))  # from (1)
        # ...
        # => var(x[i]) = var(y[i]) + var(x[i-1]) = sum[j=0,i] var(y[j])
        return torch.cumsum(t_var, dim=0)


class BBoxStandardizationTransform(InvertibleTransformWithVariance):
    """
    Applies standardization transformation:
    Y[i] = (X[i] - mean(X)) / std(X)
    """
    def __init__(self, mean: Union[float, List[float]], std: Union[float, List[float]]):
        super().__init__(name='standardization')
        self._mean = mean
        self._std = std

        if isinstance(self._mean, float):
            # Convert to list
            self._mean = [self._mean] * 4

        if isinstance(self._std, float):
            # Convert to list
            self._std = [self._std] * 4

        self._mean = torch.tensor(self._mean, dtype=torch.float32)
        self._std = torch.tensor(self._std, dtype=torch.float32)

    def apply(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        bbox_obs, bbox_unobs, *other = data
        mean = self._mean.to(bbox_obs)
        std = self._std.to(bbox_obs)

        bbox_obs = (bbox_obs - mean) / std
        # During live inference, bbox_unobs are not known (this is for training only)
        bbox_unobs = (bbox_unobs - mean) / std if bbox_unobs is not None else None

        return bbox_obs, bbox_unobs, *other

    def inverse(self, data: TensorCollection, shallow: bool = True, n_samples: int = 1) -> TensorCollection:
        bbox_obs, bbox_unobs, *other = data
        mean = self._mean.to(bbox_obs)
        std = self._std.to(bbox_obs)

        if n_samples == 1:
            # Note: inverse transform is not applied to `bbox_obs`
            bbox_unobs = bbox_unobs * std + mean
        else:
            # Support (Improvisation) for VAE monte carlo sampling for mean and std estimation
            mean_repeated = mean.repeat(n_samples)
            std_repeated = std.repeat(n_samples)
            bbox_unobs = bbox_unobs * std_repeated + mean_repeated

        return bbox_obs, bbox_unobs, *other

    def inverse_std(
        self,
        t_std: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        # Std of inverse transformation in this case is trivial to calculate
        # y - transformed, x - original
        # y[i] = (x[i] - m) / s
        # => x[i] = y[i] * s + m
        # => std(x) = std(y) * s
        std = self._std.to(t_std)
        return t_std * std

    def inverse_var(
        self,
        t_var: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        # Similar to `inverse_std`
        std = self._std.to(t_var)
        return t_var * torch.square(std)


class BBoxCompositeTransform(InvertibleTransformWithVariance):
    """
    Compose multiple transform functions.
    """
    def __init__(self, transforms: List[InvertibleTransformWithVariance]):
        """
        Args:
            transforms: List of transform functions to compose
        """
        super().__init__(name='composite')
        self._transforms = transforms

    def apply(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        for t in self._transforms:
            data = t.apply(data, shallow=shallow)
        return data

    def inverse(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        for t in self._transforms[::-1]:
            data = t.inverse(data, shallow=shallow)
        return data

    def inverse_std(
        self,
        t_std: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        for t in self._transforms[::-1]:
            t_std = t.inverse_std(t_std, additional_data=additional_data, shallow=shallow)
        return t_std

    def inverse_var(
        self,
        t_var: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        for t in self._transforms[::-1]:
            t_var = t.inverse_var(t_var, additional_data=additional_data, shallow=shallow)
        return t_var


class BBoxStandardizedFirstOrderDifferenceTransform(BBoxCompositeTransform):
    """
    This is still here for back-compatibility...

    Step 1: Applies first difference transformation:
    Y[i] = X[i] - X[i-1]
    Y[0] is removed

    Step 2: Applies standardization transformation:
    Z[i] = (Y[i] - mean(Y)) / std(Y)
    """
    def __init__(self, mean: Union[float, List[float]], std: Union[float, List[float]], scale_by_time_diff: bool = False):
        transforms = [
            BboxFirstOrderDifferenceTransform(scale_by_time_diff=scale_by_time_diff),
            BBoxStandardizationTransform(mean=mean, std=std)
        ]
        super().__init__(transforms=transforms)


class BBoxRelativeToLastObsTransform(InvertibleTransformWithVariance):
    """
    * For all observed coordinates:
    Y[i] = X[-1] - X[i]
    Y[-1] is removed

    * For all unobserved coordinates:
    Y[i] = X[i] - X[-1]
    """
    def __init__(self):
        super().__init__(name='relative_to_last_obs')

    def apply(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        bbox_obs, bbox_unobs, ts_obs, ts_unobs, *other = data
        last_obs = bbox_obs[-1:]

        if not shallow:
            bbox_obs = bbox_obs.clone()
            bbox_unobs = bbox_unobs.clone() if bbox_unobs is not None else None

        bbox_obs, ts_obs = bbox_obs[:-1], ts_obs[:-1]  # Last element becomes redundant
        ts_unobs = ts_unobs - 1  # Shifting back since the last time point is deleted
        bbox_obs = last_obs.expand_as(bbox_obs) - bbox_obs
        bbox_unobs = bbox_unobs - last_obs.expand_as(bbox_unobs) if bbox_unobs is not None else None

        return bbox_obs, bbox_unobs, ts_obs, ts_unobs, *other

    def inverse(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        orig_bbox_obs, bbox_hat, *other = data
        last_obs = orig_bbox_obs[-1:]

        if not shallow:
            bbox_hat = bbox_hat.clone()

        bbox_hat = bbox_hat + last_obs.expand_as(bbox_hat)

        return orig_bbox_obs, bbox_hat, *other

    def inverse_std(
        self,
        t_std: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        return t_std  # It's assumed that last observed element has variance equal to 0

    def inverse_var(
        self,
        t_var: torch.Tensor,
        additional_data: Optional[TensorCollection] = None,
        shallow: bool = True
    ) -> TensorCollection:
        return t_var  # It's assumed that last observed element has variance equal to 0


class BBoxStandardizedRelativeToLastObsTransform(BBoxCompositeTransform):
    """
    This is still here for back-compatibility...

    Step 1: Applies first difference transformation:
    * For all observed coordinates:
    Y[i] = X[-1] - X[i]
    Y[-1] is removed

    * For all unobserved coordinates:
    Y[i] = X[i] - X[-1]

    Step 2: Applies standardization transformation:
    Z[i] = (Y[i] - mean(Y)) / std(Y)
    """
    def __init__(self, mean: float, std: float):
        transforms = [
            BBoxRelativeToLastObsTransform(),
            BBoxStandardizationTransform(mean=mean, std=std)
        ]

        super().__init__(transforms=transforms)


class BBoxJackOfAllTradesTransform(InvertibleTransformWithVariance):
    """
    This transformation uses benefits of both relative and diff coordinates.
    Also takes into consideration the absolute object coordinates.
    """
    def __init__(
        self,
        rel_mean: float,
        rel_std: float,
        diff_mean: float,
        diff_std: float,
        rel_unobs: bool = True,
        diff_scale_by_ts_diff: bool = False
    ):
        """
        Args:
            rel_mean: Relative coordinates mean
            rel_std: Relative coordinates std
            diff_mean: Diff coordinates mean
            diff_std: Diff coordinate std
            rel_unobs: If true then Use relative coordinates for prediction
                else use diff coordinates for prediction
            diff_scale_by_ts_diff: Scale feature difference by time difference
        """
        super().__init__(name='jack_of_all_trades')
        self._rel_transform = BBoxStandardizedRelativeToLastObsTransform(mean=rel_mean, std=rel_std)
        self._diff_transform = BBoxStandardizedFirstOrderDifferenceTransform(mean=diff_mean, std=diff_std, scale_by_time_diff=diff_scale_by_ts_diff)
        self._rel_unobs = rel_unobs

    def apply(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        bbox_obs, bbox_unobs, ts_obs, ts_unobs, *other = data
        rel_bbox_obs, rel_bbox_unobs, rel_ts_obs, rel_ts_unobs, *_ = \
            self._rel_transform([bbox_obs, bbox_unobs, ts_obs, ts_unobs, None], shallow=False)
        diff_bbox_obs, diff_bbox_unobs, *_ = self._diff_transform([bbox_obs, bbox_unobs, ts_obs, None], shallow=False)
        bbox_obs = bbox_obs[1:]
        bbox_features = torch.cat([rel_bbox_obs, diff_bbox_obs, bbox_obs], dim=-1)

        bbox_unobs = rel_bbox_unobs if self._rel_unobs else diff_bbox_unobs
        return bbox_features, bbox_unobs, rel_ts_obs, rel_ts_unobs, *other

    def inverse(self, data: TensorCollection, shallow: bool = True) -> TensorCollection:
        if self._rel_unobs:
            return self._rel_transform.inverse(data, shallow=shallow)
        return self._diff_transform.inverse(data, shallow=shallow)

    def inverse_std(self, t_std: torch.Tensor, additional_data: Optional[TensorCollection] = None, shallow: bool = True) -> TensorCollection:
        if self._rel_unobs:
            return self._rel_transform.inverse_std(t_std, additional_data=additional_data, shallow=shallow)
        return self._diff_transform.inverse_std(t_std, additional_data=additional_data, shallow=shallow)

    def inverse_var(self, t_var: torch.Tensor, additional_data: Optional[TensorCollection] = None, shallow: bool = True) -> TensorCollection:
        if self._rel_unobs:
            return self._rel_transform.inverse_var(t_var, additional_data=additional_data, shallow=shallow)
        return self._diff_transform.inverse_var(t_var, additional_data=additional_data, shallow=shallow)


# noinspection DuplicatedCode
def run_test_first_difference() -> None:
    bbox_obs = torch.randn(2, 2, 4)
    bbox_unobs = torch.randn(3, 2, 4)
    ts_obs = torch.tensor([1, 3], dtype=torch.float32).view(-1, 1, 1).repeat(1, 2, 1)
    ts_unobs = torch.tensor([4, 5, 6], dtype=torch.float32).view(-1, 1, 1).repeat(1, 2, 1)

    for scale_by_time_diff in [False, True]:
        first_diff = BboxFirstOrderDifferenceTransform(scale_by_time_diff=scale_by_time_diff)

        transformed_bbox_obs, transformed_bbox_unobs, *_ = \
            first_diff.apply([bbox_obs, bbox_unobs, ts_obs, ts_unobs, None], shallow=False)
        print(transformed_bbox_obs)
        assert transformed_bbox_obs.shape == (1, 2, 4)
        assert transformed_bbox_unobs.shape == (3, 2, 4)

        _, inv_transformed_bbox_unobs, *_ = first_diff.inverse([bbox_obs, transformed_bbox_unobs, None])
        assert torch.abs(inv_transformed_bbox_unobs - bbox_unobs).sum().item() < 1e-3


def run_test_standardization() -> None:
    bbox_obs = torch.randn(2, 2, 4)
    bbox_unobs = torch.randn(3, 2, 4)
    ts_obs = torch.randn(2, 2, 1)
    ts_unobs = torch.randn(3, 2, 1)
    first_diff = BBoxStandardizationTransform(1, 2.0)

    _, transformed_bbox_unobs, *_ = \
        first_diff.apply([bbox_obs, bbox_unobs, ts_obs, ts_unobs, None], shallow=False)

    _, inv_transformed_bbox_unobs, *_ = first_diff.inverse([bbox_obs, transformed_bbox_unobs, None])
    assert torch.abs(inv_transformed_bbox_unobs - bbox_unobs).sum().item() < 1e-3


# noinspection DuplicatedCode
def run_test_standardized_first_difference() -> None:
    bbox_obs = torch.randn(2, 2, 4)
    bbox_unobs = torch.randn(3, 2, 4)
    ts_obs = torch.randn(2, 2, 1)
    ts_unobs = torch.randn(3, 2, 1)
    first_diff = BBoxStandardizedFirstOrderDifferenceTransform(1, 2.0)

    transformed_bbox_obs, transformed_bbox_unobs, *_ = \
        first_diff.apply([bbox_obs, bbox_unobs, ts_obs, ts_unobs, None], shallow=False)
    assert transformed_bbox_obs.shape == (1, 2, 4)
    assert transformed_bbox_unobs.shape == (3, 2, 4)

    _, inv_transformed_bbox_unobs, *_ = first_diff.inverse([bbox_obs, transformed_bbox_unobs, None])
    assert torch.abs(inv_transformed_bbox_unobs - bbox_unobs).sum().item() < 1e-3


# noinspection DuplicatedCode
def run_test_relative_to_last_obs() -> None:
    bbox_obs = torch.randn(2, 2, 4)
    bbox_unobs = torch.randn(3, 2, 4)
    ts_obs = torch.randn(2, 2, 1)
    ts_unobs = torch.randn(3, 2, 1)
    first_diff = BBoxRelativeToLastObsTransform()

    transformed_bbox_obs, transformed_bbox_unobs, *_ = \
        first_diff.apply([bbox_obs, bbox_unobs, ts_obs, ts_unobs, None], shallow=False)
    assert transformed_bbox_obs.shape == (1, 2, 4)
    assert transformed_bbox_unobs.shape == (3, 2, 4)

    _, inv_transformed_bbox_unobs, *_ = first_diff.inverse([bbox_obs, transformed_bbox_unobs, None])
    assert torch.abs(inv_transformed_bbox_unobs - bbox_unobs).sum().item() < 1e-3


# noinspection DuplicatedCode
def run_test_standardized_relative_to_last_obs() -> None:
    bbox_obs = torch.randn(2, 2, 4)
    bbox_unobs = torch.randn(3, 2, 4)
    ts_obs = torch.randn(2, 2, 1)
    ts_unobs = torch.randn(3, 2, 1)
    first_diff = BBoxStandardizedRelativeToLastObsTransform(1, 2.0)

    transformed_bbox_obs, transformed_bbox_unobs, *_ = \
        first_diff.apply([bbox_obs, bbox_unobs, ts_obs, ts_unobs, None], shallow=False)
    assert transformed_bbox_obs.shape == (1, 2, 4)
    assert transformed_bbox_unobs.shape == (3, 2, 4)

    _, inv_transformed_bbox_unobs, *_ = first_diff.inverse([bbox_obs, transformed_bbox_unobs, None])
    assert torch.abs(inv_transformed_bbox_unobs - bbox_unobs).sum().item() < 1e-3


def run_test_jack_of_all_trades() -> None:
    for use_rel_unobs in [True, False]:
        bbox_obs = torch.randn(2, 2, 4)
        bbox_unobs = torch.randn(3, 2, 4)
        ts_obs = torch.randn(2, 2, 1)
        ts_unobs = torch.randn(3, 2, 1)
        jack_of_all_trades = BBoxJackOfAllTradesTransform(1.0, 2.0, 1.0, 2.0, rel_unobs=use_rel_unobs)

        transformed_bbox_obs, transformed_bbox_unobs, *_ = \
            jack_of_all_trades.apply([bbox_obs, bbox_unobs, ts_obs, ts_unobs, None], shallow=False)
        assert transformed_bbox_obs.shape == (1, 2, 12)
        assert transformed_bbox_unobs.shape == (3, 2, 4)

        _, inv_transformed_bbox_unobs, *_ = jack_of_all_trades.inverse([bbox_obs, transformed_bbox_unobs, None])
        assert torch.abs(inv_transformed_bbox_unobs - bbox_unobs).sum().item() < 1e-3


if __name__ == '__main__':
    run_test_first_difference()
    run_test_standardization()
    run_test_standardized_first_difference()
    run_test_relative_to_last_obs()
    run_test_standardized_relative_to_last_obs()
    run_test_jack_of_all_trades()

    # TODO: Move to tests



"""
MOT/SOT datasets
"""
from motrack_motion.datasets import transforms
from motrack_motion.datasets.factory import dataset_factory, DATASET_CATALOG
from motrack_motion.datasets.torch import TrajectoryDataset, TorchTrajectoryDataset
from motrack_motion.datasets.mot.core import MOTDataset
from motrack_motion.datasets.utils import TrajectoryDataloaderCollateFunctional



"""
Scene basic information.
"""
from dataclasses import dataclass
from typing import Union


@dataclass
class BasicSceneInfo:
    """
    Basic scene information
    """
    name: str
    category: str
    seqlength: Union[int, str]  # Can be parsed from `str`
    imheight: Union[int, str]
    imwidth: Union[int, str]

    def __post_init__(self):
        """
        Convert to proper type.
        """
        self.seqlength = int(self.seqlength)
        self.imheight = int(self.imheight)
        self.imwidth = int(self.imwidth)
